/* Copyright (C) 1994, 1995 Aladdin Enterprises.  All rights reserved.
  
  This file is part of Aladdin Ghostscript.
  
  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
  or distributor accepts any responsibility for the consequences of using it,
  or for whether it serves any particular purpose or works at all, unless he
  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
  License (the "License") for full details.
  
  Every copy of Aladdin Ghostscript must include a copy of the License,
  normally in a plain ASCII text file named PUBLIC.  The License grants you
  the right to copy, modify and redistribute Aladdin Ghostscript, but only
  under certain conditions described in the License.  Among other things, the
  License requires that the copyright notice and this notice be preserved on
  all copies.
*/


#include "MacHeaders"
#ifdef WIN_PALLETES
#include "Palettes.h"
#endif
#include "QDOffscreen.h"
#include <stdio.h>
#include <stddef.h>
#include <string.h>
#include "gserrors.h"
#include "gdevmac.h"
#include "gsparam.h"
#include "gp_mac.h"


/* Device Procedures */

private dev_proc_open_device(mac_open);
private dev_proc_get_initial_matrix(mac_get_initial_matrix);
private dev_proc_sync_output(mac_sync);
private dev_proc_output_page(mac_output_page);
private dev_proc_get_params(mac_get_params);
private dev_proc_put_params(mac_put_params);
private dev_proc_close_device(mac_close);
private dev_proc_map_rgb_color(mac_bw_rgb_color);
private dev_proc_map_color_rgb(mac_bw_color_rgb);
private dev_proc_map_rgb_color(mac_gray_rgb_color);
private dev_proc_map_color_rgb(mac_gray_color_rgb);
private dev_proc_map_rgb_color(mac_8bit_color_rgb_color);
private dev_proc_map_color_rgb(mac_8bit_color_color_rgb);
private dev_proc_fill_rectangle(mac_fill_rectangle);
private dev_proc_tile_rectangle(mac_tile_rectangle);
private dev_proc_copy_mono(mac_copy_mono);
private dev_proc_copy_color(mac_copy_color);
private dev_proc_draw_line(mac_draw_line);

//#define COLOR_TEST

/* The device descriptor */

/* in the future we should call gx_default_put_params_only to 
	do the work of writing the page parameters and then adjust 
	ourselves accordingly */

gx_device_procs gs_mac_procs = {
	mac_open,
	mac_get_initial_matrix,
	mac_sync,
	mac_output_page,
	mac_close,
	mac_bw_rgb_color,
	mac_bw_color_rgb,
	mac_fill_rectangle,
	mac_tile_rectangle,
	mac_copy_mono,
	mac_copy_color,
	mac_draw_line,
	gx_default_get_bits,
	mac_get_params,
	mac_put_params,
	gx_default_map_cmyk_color,
	gx_default_get_xfont_procs,
	gx_default_get_xfont_device,
	gx_default_map_rgb_alpha_color,
	gx_default_get_page_device
};
	
/* The instance is public. */

gx_device_mac gs_mac_device =
{
	std_device_color_body(
		gx_device_mac,
		&gs_mac_procs,
		DEV_MAC_NAME,
 		DEFAULT_DEV_WIDTH, DEFAULT_DEV_HEIGHT,	/* x and y extent */
 		DEFAULT_DEV_DPI, DEFAULT_DEV_DPI,		/* x and y density */
		1, 1, 2),								/* default to b & w */
	0,								/* not shown on open */
	{ 0 },							/* no title */
	0								/* color table not filled in */
};




/* Macros for colors */

#define WHITE_INDEX		((gx_color_index)0)

#ifndef WIN_PALETTE
#define xIndexForeColor(c)								\
	if (dp->lastForeColor != c) {					\
		RGBForeColor(&(mdev->colors[c].rgb));		\
		dp->lastForeColor = c;						\
	}
#define IndexForeColor(c)	RGBForeColor(&(mdev->colors[c].rgb))
#define IndexBackColor(c)	RGBBackColor(&(mdev->colors[c].rgb))
#endif



/* Macro to validate arguments */

#define check_rect()\
	if ( w <= 0 || h <= 0 ) return 0;\
	if ( x < 0 || x > dev->width - w || y < 0 || y > dev->height - h )\
		return_error (gs_error_rangecheck)


#define BEGIN_DRAW										\
preDrawCheck(mdev);										\
if (mdev->mwin != NULL) {								\
	CGrafPtr 	savePort;								\
	GDHandle	saveDev;								\
	draw_port	* dp;									\
														\
	GetGWorld(&savePort, &saveDev);						\
														\
	for (dp = &mdev->mwin->win; dp != NULL; dp = dp->next) {\
		setUpDrawPort(dp);								\
		{

#define END_DRAW										\
		}												\
		takeDownDrawPort(dp);							\
		SetGWorld(savePort, saveDev);					\
	}													\
}


private void
setUpDrawPort(draw_port * dp)
{
	if (dp->isGWorld)
		SetGWorld(dp->port, nil);
	else
		SetPort((GrafPtr)dp->port);
	
	if (dp->setOriginAndClip) {
		Rect		viewRect;
		
		viewRect = dp->viewRect;
		OffsetRect(&viewRect, dp->origin.h, dp->origin.v);
		SetOrigin(dp->origin.h, dp->origin.v);
		ClipRect(&viewRect);
	}
}

private void
takeDownDrawPort(draw_port * dp)
{
	if (dp->setOriginAndClip) {
		SetOrigin(0, 0);
		ClipRect(&dp->port->portRect);
	}
	
	//ForeColor(blackColor);
	//BackColor(whiteColor);
}



private void
preDrawCheck(gx_device_mac * mdev)
{
	mac_gs_win		* mwin;
	
	mwin = mdev->mwin;
	
	if (mwin != NULL
	&&  mwin->pageOutput) {
		currWindow = mwin;
		confirm_next_page(mwin);
		mwin = mdev->mwin;	// could change mwin...
	}
	
	// Automatically create the window if it's missing
	// and we're not in an interrupt
	if (mwin == NULL  &&  ! interruptRequest) {
		mwin = macw_create_window(mdev);
		
		if (mwin == NULL) {
			// create failed, might as well kill this render...
			interruptRequest = true;
			return;
		}
	}

	if (mwin != NULL) {
		if (mwin->delayedClear)
			macw_clear_window(mwin);
		
		mwin->pageOutput = false;
		mwin->delayedClear = false;
		currWindow = mwin;
	}
}






private int
mac_open(register gx_device *dev)
{
	gx_device_mac			* mdev = (gx_device_mac *)dev;
	gx_device_color_info	* ci;

	/* fix up color info */
	if (mac_8bit_clut == NULL) init_palette();

	ci = &mdev->color_info;
	switch (ci->depth) {
		case 1:
			ci->num_components = 1;
			ci->max_gray = 1;		ci->max_color = 0;
			ci->dither_grays = 2;	ci->dither_colors = 0;
			mdev->clut = bw_clut;
			mdev->colors = bw_colors;
			set_dev_proc(dev, map_rgb_color, mac_bw_rgb_color);
			set_dev_proc(dev, map_color_rgb, mac_bw_color_rgb);
			break;
			
		case 4:
			ci->num_components = 1;
			ci->max_gray = 15;		ci->max_color = 0;
			ci->dither_grays = 16;	ci->dither_colors = 0;
			mdev->clut = gray_4bit_clut;
			mdev->colors = gray_4bit_colors;
			set_dev_proc(dev, map_rgb_color, mac_gray_rgb_color);
			set_dev_proc(dev, map_color_rgb, mac_gray_color_rgb);
			break;
		
		default:
		case 8:
			ci->depth = 8;	// just in case it was anything else
			ci->num_components = 3;
			ci->max_gray = 15;		ci->max_color = 5;
			ci->dither_grays = 16;	ci->dither_colors = 6;
			mdev->clut = mac_8bit_clut;
			mdev->colors = mac_8bit_colors;
			set_dev_proc(dev, map_rgb_color, mac_8bit_color_rgb_color);
			set_dev_proc(dev, map_color_rgb, mac_8bit_color_color_rgb);
			break;
	}

#if MAC_DEV_DEBUG
	{
		char	message[100];
		
		sprintf(message, "Opening device <%lx>.\n", (long)dev);	
		zcPuts(theConsole, message);
	}
#endif
	mdev->mwin = NULL;		// may not be null since device could have been copied.
	
	if (mdev->visible)
		macw_create_window(mdev);
	
	return 0;
}


/* Close the device. */
private int
mac_close(register gx_device *dev)
{
	gx_device_mac		* mdev = (gx_device_mac *)dev;
	mac_gs_win			* mwin = mdev->mwin;
#if MAC_DEV_DEBUG
	{
		char	message[100];
		
		sprintf(message, "Closing device <%lx>.\n", (long)dev);	
		zcPuts(theConsole, message);
	}
#endif	
	if (mwin != NULL) {
		macw_detach_window(mwin);
	}
	
	return 0;
}



/* Fill a rectangle with a color. */
private int
mac_fill_rectangle(register gx_device *dev,
	int x, int y, int w, int h, gx_color_index color)
{
	gx_device_mac		* mdev = (gx_device_mac *)dev;
	mac_gs_win			* mwin = mdev->mwin;
	
	check_rect();
	if (color == gx_no_color_index) return 0;
	
	if (x == 0  &&  y == 0
	&&  w == dev->width  &&  h == dev->height
	&&  color == WHITE_INDEX) {
		// this is a clear
		if (mwin == NULL) {
			// no window, so nothing to clear:
			// the new window will be clear when created
			return 0;
		}
		else if (mwin->pageOutput) {
			// if there is a window and it is a compleated page,
			// delay the clear 'till later
		  	mwin->delayedClear = true;
			return 0;
		}
		// let the clear go through...
	}
	
	if (w == 1  &&  h == 1) {
		BEGIN_DRAW
	    	SetCPixel(x, y, &(mdev->colors[color].rgb));
		END_DRAW
	}
	else {
		Rect		theRect;
	
	    SetRect(&theRect, x, y, x + w, y + h);
	    
		BEGIN_DRAW
			IndexForeColor(color);
	    	PaintRect(&theRect);
		END_DRAW
	}
	
	return 0;
}


/* tile a rectangle with a bitmap or pixmap */
private int
mac_tile_rectangle(register gx_device *dev,
	const gx_tile_bitmap *tile,
	int x, int y, int w, int h,
	gx_color_index color_0, gx_color_index color_1,
	int phase_x, int phase_y)
{
	gx_device_mac		* mdev = (gx_device_mac *)dev;
	mac_gs_win			* mwin = mdev->mwin;
	int					ix, fx, i;
	int					iy, fy, j;
	Rect				oldClip;
	
	if (color_0 == gx_no_color_index
	&&  color_1 == gx_no_color_index)
		goto doDefault;

	check_rect();
	
	ix = (x + phase_x) / tile->size.x;
	iy = (y + phase_y) / tile->size.y;
	fx = (x + w + phase_x + tile->size.x - 1) / tile->size.x;
	fy = (y + h + phase_y + tile->size.y - 1) / tile->size.y;
	
	BEGIN_DRAW
		Rect		srcRect, dstRect;
		BitMap		srcBits, *dstBits;
		short		mode;
		
		mode = srcCopy;
		
		if (color_0 != gx_no_color_index) {
			IndexBackColor(color_0);
		}
		else {
			mode = srcOr;
		}

		if (color_1 != gx_no_color_index) {
			IndexForeColor(color_1);
		}
		else {
			mode = notSrcBic;
		}

	    SetRect(&dstRect, x, y, x + w, y + h);
	    oldClip = (*dp->port->clipRgn)->rgnBBox;
	    SectRect(&oldClip, &dstRect, &dstRect);
	    ClipRect(&dstRect);
	    
	    SetRect(&srcRect, 0, 0, tile->size.x, tile->size.y);
	    srcBits.baseAddr	= (QDPtr) tile->data;
	    srcBits.rowBytes	= tile->raster;
	    srcBits.bounds		= srcRect;
	    dstBits = (BitMap *)&dp->port->portPixMap;

		for (i = ix; i < fx; i += 1) {
			for (j = iy; j < fy; j += 1) {
	   			dstRect = srcRect;
	   			OffsetRect(&dstRect,
	   				i * tile->size.x - phase_x,
	   				j * tile->size.y - phase_y);

				CopyBits(&srcBits, dstBits, &srcRect, &dstRect, mode, NULL);
			}
		}
		
		ClipRect(&oldClip);
	END_DRAW

	return 0;

doDefault:
	return gx_default_tile_rectangle(dev, tile,
		x, y, w, h,
		color_0, color_1, phase_x, phase_y);
}


/* Copy a monochrome bitmap. */
private int
mac_copy_mono (register gx_device *dev,
			   const unsigned char *base, int data_x, int raster, gx_bitmap_id id,
			   int x, int y, int w, int h,
			   gx_color_index color_0, gx_color_index color_1)
{
	gx_device_mac		* mdev = (gx_device_mac *)dev;
	mac_gs_win			* mwin = mdev->mwin;
	
	check_rect();

	BEGIN_DRAW
		Rect		srcRect, dstRect;
		BitMap		srcBits, *dstBits;
		short		mode;
		
		mode = srcCopy;
		
		if (color_0 != gx_no_color_index) {
			IndexBackColor(color_0);
		}
		else {
			mode = srcOr;
		}

		if (color_1 != gx_no_color_index) {
			IndexForeColor(color_1);
		}
		else {
			if (mode == srcOr)
				mode = -1;
			else
				mode = notSrcBic;
		}

	    SetRect(&srcRect, data_x, 0, data_x + w, h);
	    SetRect(&dstRect, x, y, x + w, y + h);
	    srcBits.baseAddr	  = (QDPtr) base;
	    srcBits.rowBytes	  = raster;
	    srcBits.bounds.left	  = 0;
	    srcBits.bounds.top    = 0;
	    srcBits.bounds.right  = raster * 8;
	    srcBits.bounds.bottom = h;
	    dstBits = (BitMap *)&dp->port->portPixMap;

		if (mode != -1)
			CopyBits(&srcBits, dstBits, &srcRect, &dstRect, mode, NULL);
	END_DRAW

	return 0;
}


/* Copy a color bitmap. */
private int
mac_copy_color (register gx_device *dev,
				const unsigned char *base, int sourcex, int raster,  gx_bitmap_id id,
				int x, int y, int w, int h)
{
	gx_device_mac		* mdev = (gx_device_mac *)dev;
	mac_gs_win			* mwin = mdev->mwin;
	Rect				srcRect, dstRect;
	PixMap				*srcBits;
	static PixMapHandle	pmaph = NULL;
		
	check_rect();
	
	if (pmaph == NULL) {
		pmaph = NewPixMap();
		if (pmaph == NULL) noMemoryExit();
		MoveHHi((Handle)pmaph);
		HLock((Handle)pmaph);
		
	    (*pmaph)->pixelType		= clutType;
		(*pmaph)->pixelSize		= dev->color_info.depth;
		(*pmaph)->cmpCount		= 1;
		(*pmaph)->cmpSize		= dev->color_info.depth;
		(*pmaph)->planeBytes	= 0;
		(*pmaph)->pmTable		= mdev->clut;
	}

    SetRect(&srcRect, sourcex, 0, sourcex + w, h);
    SetRect(&dstRect, x, y, x + w, y + h);
    srcBits = *pmaph;
    srcBits->baseAddr		= (QDPtr) base;
    srcBits->rowBytes		= raster + 0x8000;
    srcBits->bounds.left	= 0;
    srcBits->bounds.top		= 0;
    srcBits->bounds.right	= raster;
    srcBits->bounds.bottom	= h;
    
	BEGIN_DRAW
		CopyBits((BitMap *)srcBits, (BitMap *)&dp->port->portPixMap,
				 &srcRect,			&dstRect,
				 srcCopy, NULL);
	END_DRAW

	return 0;
}


/* Draw a line */

private int
mac_draw_line (register gx_device *dev,
			   int x0, int y0,
			   int x1, int y1,
			   gx_color_index color)
{
	gx_device_mac		* mdev = (gx_device_mac *)dev;
	mac_gs_win			* mwin = mdev->mwin;
	
	if (color == gx_no_color_index) return 0;
	if (x0 == x1  &&  y0 == y1) return 0;
	if (y1 < y0) {
		int		t;
		t = x0; x0 = x1; x1 = t;
		t = y0; y0 = y1; y1 = t;
	}
	
	// adjust for pen, this still isn't right!
	//if (y1 > y0)	y1 -= 1;
	//if (x1 > x0)	x1 -= 1;
	
	BEGIN_DRAW
		IndexForeColor(color);

		MoveTo(x0, y0);
		LineTo(x1, y1);
	END_DRAW

	return 0;
}




private gx_color_index
cube_color_index(register uint r,register  uint g,register  uint b)
{
	// convert a color in 6,6,6 space to the cube index
	register gx_color_index		c;
	
	c = (r * 36 + g * 6 + b);
	return (c > 0) ? 215 - c : 255;		// black is at end
}

private gx_color_index
ramp_color_index(register gx_color_index offset, register uint v)
{
	// convert a color in 16,16,16 space to the ramp
	register uint		w;
	
	w = v % 3;
	v = (v - w) / 3;
	
	if (w == 0) {
		switch (offset) {
			case 215:	return cube_color_index(v, 0, 0);
			case 225:	return cube_color_index(0, v, 0);
			case 235:	return cube_color_index(0, 0, v);
			case 245:	return cube_color_index(v, v, v);
		}
	}
	
	return offset + 10 - (v * 2 + w);
}

private gx_color_index
mac_8bit_color_rgb_color(gx_device *dev, ushort r, ushort g, ushort b)
{
	register uint r16, g16, b16;
	
	// find out where we are in 16,16,16 space, (which is 3x 6,6,6 space)
	r16 = r / (gx_max_color_value / 15);
	g16 = g / (gx_max_color_value / 15);
	b16 = b / (gx_max_color_value / 15);
		
	// check for balck
	if (r16 == 0  &&  g16 == 0  &&  b16 == 0)
		return 255;

	// check for being in the gray ramp
	if (r16 == g16  &&  g16 == b16)	return ramp_color_index(245, r16);
	
	// check for being in the color ramps
	if (g16 <= 1  &&  b16 <= 1)	return ramp_color_index(215, r16);
	if (r16 <= 1  &&  b16 <= 1)	return ramp_color_index(225, g16);
	if (r16 <= 1  &&  g16 <= 1)	return ramp_color_index(235, b16);

	// else we're in the cube
	return cube_color_index(r16 / 3, g16 / 3, b16 / 3);
}


private int
mac_8bit_color_color_rgb (gx_device *dev,
				   register gx_color_index color, ushort prgb[3])
{
	if (color < 215) {
		// we are in strict 6,6,6 space
		ushort		r, g, b;
		
		b = color % 6;
		g = ((color - b) / 6) % 6;
		r = (color - b - 6 * g) / 36;
		
		prgb[0] = (5 - r) * (gx_max_color_value / 5);
		prgb[1] = (5 - g) * (gx_max_color_value / 5);
		prgb[2] = (5 - b) * (gx_max_color_value / 5);
	}
	else if (color < 255) {
		// we are in the ramps
		ushort		rmp, v, w;

		color -= 215;
		v = color % 10;
		rmp = (color - v) / 10;
		
		w = v % 2;
		v = (v - w) / 2;
		v = ((gx_max_color_value / 5) * (5 - v))
		  - ((gx_max_color_value / 15) * (2 - w));
		 
		switch (rmp) {
			case 0:	prgb[0] = v; prgb[1] = 0; prgb[2] = 0; break;
			case 1:	prgb[0] = 0; prgb[1] = v; prgb[2] = 0; break;
			case 2:	prgb[0] = 0; prgb[1] = 0; prgb[2] = v; break;
			case 3:	prgb[0] = v; prgb[1] = v; prgb[2] = v; break;
		}
	}
	else {
		// we are black
		prgb[0] = 0; prgb[1] = 0; prgb[2] = 0;
	}
	
	return 0;
}

private gx_color_index
mac_gray_rgb_color(gx_device *dev, ushort r, ushort g, ushort b)
{
	return dev->color_info.max_gray - gx_default_gray_map_rgb_color(dev, r, g, b);
}

private int
mac_gray_color_rgb (gx_device *dev,
				   register gx_color_index color, ushort prgb[3])
{
	return gx_default_gray_map_color_rgb(dev, dev->color_info.max_gray - color, prgb);
}

private gx_color_index
mac_bw_rgb_color(gx_device *dev, ushort r, ushort g, ushort b)
{
	return 1 - gx_default_map_rgb_color(dev, r, g, b);
}

private int
mac_bw_color_rgb (gx_device *dev,
				   register gx_color_index color, ushort prgb[3])
{
	return gx_default_map_color_rgb(dev, 1 - color, prgb);
}


/* Between output and sync, if this is a showpage, we will receive
   a call to clear the page (a fill_rectangle).  In this case, we
   simply retain that fact and do the clear on the next graphic
   operation.  That way we don't clear the image on compleation. */
   
private int
mac_output_page (gx_device *dev, int num_copies, int flush)
{
	gx_device_mac		* mdev = (gx_device_mac *)dev;
	mac_gs_win			* mwin = mdev->mwin;
	
	if (mwin != NULL)
		mwin->pageOutput = true;
		
	return 0;
}

private int
mac_sync(register gx_device *dev)
{
	return 0;
}


private int
mac_get_params(gx_device *dev, gs_param_list *plist)
{
	gx_device_mac		* mdev = (gx_device_mac *)dev;
	mac_gs_win			* mwin = mdev->mwin;
	int					visible;
	int					depth;
	gs_param_string		title;
	int					code;
	
	if (!mdev->is_open)
		mwin = NULL;
	
	if (mwin == NULL) {
		visible = mdev->visible;
	}
	else {
		visible = ((WindowPeek)mwin->win.port)->visible;
		GetWTitle((WindowPtr)mwin->win.port, mdev->title);
	}
	title.data = &mdev->title[1];
	title.size = mdev->title[0];
	title.persistent = false;
	
	depth = dev->color_info.depth;
	
	code = gx_default_get_params(dev, plist);
	if (code < 0) return code;
	
	code = param_write_bool(plist, "GSVisible", &visible);
	if (code < 0) return code;
	
	code = param_write_string(plist, "GSTitle", &title);
	if (code < 0) return code;
	
	code = param_write_int(plist, "BitsPerPixel", &depth);
	return code;
}

private int
put_param_bool(gs_param_list *plist, gs_param_name pname, int *pvalue, int ecode)
{
	int code, value;

	switch ( code = param_read_bool(plist, pname, &value) ) {
	default:	return code;
	case 1:		return ecode;
	case 0:
		*pvalue = value;
		return (ecode < 0 ? ecode : 1);
	}
}

private int
put_param_int(gs_param_list *plist, gs_param_name pname, int *pvalue, int ecode)
{
	int code, value;

	switch ( code = param_read_int(plist, pname, &value) ) {
	default:	return code;
	case 1:		return ecode;
	case 0:
		*pvalue = value;
		return (ecode < 0 ? ecode : 1);
	}
}

private int
put_param_string(gs_param_list *plist, gs_param_name pname,
	gs_param_string *pvalue, int ecode)
{
	int					code;
	gs_param_string		value;

	switch ( code = param_read_string(plist, pname, &value) ) {
	default:	return code;
	case 1:		return ecode;
	case 0:
		*pvalue = value;
		return (ecode < 0 ? ecode : 1);
	}
}	


private int
mac_put_params(gx_device *dev, gs_param_list *plist)
{	
	gx_device_mac		* mdev = (gx_device_mac *)dev;
	mac_gs_win			* mwin = mdev->mwin;
	int					visible;
	int					depth, currDepth;
	gs_param_string		title;
	int					code = 0;
	
	if (!mdev->is_open)
		mwin = NULL;
	
	if (mwin == NULL)
		visible = mdev->visible;
	else
		visible = ((WindowPeek)mwin->win.port)->visible;
	
	title.data = NULL;	// flag that it wasn't set
	depth = mdev->color_info.depth;
	
	code = put_param_bool(plist, "GSVisible", &visible, code);
	code = put_param_string(plist, "GSTitle", &title, code);
	code = put_param_int(plist, "BitsPerPixel", &depth, code);
	if (depth != 1  &&  depth != 4  && depth != 8)
		param_return_error(plist, "BitsPerPixel", gs_error_rangecheck);
	
	if ( code < 0 )	return code;
	currDepth = mdev->color_info.depth;	// fool gx_default_put_params to not complain
	mdev->color_info.depth = depth;
	code = gx_default_put_params(dev, plist);
	mdev->color_info.depth = currDepth;	// put it back so any failures won't change it
	if ( code < 0 )	return code;

#if MAC_DEV_DEBUG
	{
		char	message[200];
		
		sprintf(message, "Putting params for device <%lx>.\n", (long)dev);	
		zcPuts(theConsole, message);
		sprintf(message, "\tPage = [%f, %f]\n\tDevSize = [ %d, %d ]\n\tResolution = [ %f, %f ]\n",
			dev->PageSize[0], dev->PageSize[1],
			dev->width, dev->height,
			dev->HWResolution[0], dev->HWResolution[1]);
		zcPuts(theConsole, message); 
	}
#endif
	if (title.data != NULL) {
		if (title.size > 255) title.size = 255;
		mdev->title[0] = title.size;
		memcpy(&mdev->title[1], title.data, title.size);
	}
	
	mdev->visible = visible;
	
	if (mwin != NULL) {
		SetWTitle((WindowPtr)mwin->win.port, mdev->title);
		if (visible)
			ShowWindow((WindowPtr)mwin->win.port);
		else
			HideWindow((WindowPtr)mwin->win.port);
	}
	
	if (depth != mdev->color_info.depth) {
		if (dev->is_open) gs_closedevice(dev);
		mdev->color_info.depth = depth;
		// open will reset all color things correctly
	}

	return code;
}

private void
mac_get_initial_matrix (register gx_device *dev,
						register gs_matrix *pmat)
{
	pmat->xx = dev->x_pixels_per_inch /  72.0;
	pmat->xy = 0;
	pmat->yx = 0;
	pmat->yy = dev->y_pixels_per_inch / -72.0;
	pmat->tx = 0;
	pmat->ty = dev->height;
}
