/* Copyright (C) 1989 - 1995 Aladdin Enterprises.  All rights reserved.
  
  This file is part of Aladdin Ghostscript.
  
  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
  or distributor accepts any responsibility for the consequences of using it,
  or for whether it serves any particular purpose or works at all, unless he
  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
  License (the "License") for full details.
  
  Every copy of Aladdin Ghostscript must include a copy of the License,
  normally in a plain ASCII text file named PUBLIC.  The License grants you
  the right to copy, modify and redistribute Aladdin Ghostscript, but only
  under certain conditions described in the License.  Among other things, the
  License requires that the copyright notice and this notice be preserved on
  all copies.
*/

#include "MacHeaders"

#include "memory_.h"
#include "string_.h"
#include "gx.h"
#include "gp.h"
#include "gpcheck.h"
#include "gp_mac.h"
#include "gdebug.h"
#include "sys/stat.h"
#include <stdarg.h>





/* Do platform-dependent initialization */

void
gp_init (void)

{
	extern char    *gs_lib_default_path;
	extern char    *gs_init_file;
	
//	__open_std ();

#if 0
	/*...Initialize Ghostscript's default library paths and initialization file...*/

	{
		int			i;
		char	  **p;


		for (i = iGSLibPathStr, p = &gs_lib_default_path;
			 i <= iGSInitFileStr;
			 i++, p = &gs_init_file)
		{
			GetIndString (string, MACSTRS_RES_ID, i);
			(void) PtoCstr (string);
			*p = gs_malloc (1, (size_t) (strlen ((char *) string) + 1), "gp_init");
			strcpy (*p, (char *) string);
		}
	}
#endif
}


/* Do platform-dependent cleanup. */


void
gp_exit (int exit_status, int code)
{
	extern void exitAll(void);
	
	exitAll();
}



/*
 * Get the string corresponding to an OS error number.
 * If no string is available, return NULL.  The caller may assume
 * the string is allocated statically and permanently.
 */
const char *	gp_strerror(P1(int))
{
	return NULL;
}


/* Read the current date (in days since Jan. 1, 1980) */
/* and time (in milliseconds since midnight). */

	void
gp_get_clock (long *pdt)

{
	long				secs;
	DateTimeRec			dateRec;
	static DateTimeRec	baseDateRec = {1980, 1, 1, 0, 0, 0, 1};
	long				pdtmp[2];
	void 				do_get_clock (DateTimeRec *dateRec, long *pdt);


	GetDateTime ((unsigned long *) &secs);
	Secs2Date (secs, &dateRec);

	do_get_clock (&dateRec	  , pdt);
	do_get_clock (&baseDateRec, pdtmp);

	/* If the date is reasonable, subtract the days since Jan. 1, 1980 */

	if (pdtmp[0] < pdt[0])
		pdt[0] -= pdtmp[0];

#ifdef DEBUG_CLOCK
	printf("pdt[0] = %ld  pdt[1] = %ld\n", pdt[0], pdt[1]);
#endif
}


	static void
do_get_clock (DateTimeRec *dateRec, long *pdt)

{
	long 	idate;
	static int mstart[12] =
	   {0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334};


	idate = (long) dateRec->year * 365 +
		(dateRec->year / 4 + 1 +		/* account for leap years */
		 mstart[dateRec->month - 1] +	/* month is 1-origin */
		 dateRec->day - 1);				/* day of month is 1-origin */
	if (dateRec->month <= 2 && dateRec->year % 4 == 0)		/* Jan. or Feb. of leap year */
		idate--;

	pdt[0] = idate;
	pdt[1] = (dateRec->hour * 60L + dateRec->minute) * 60000L + dateRec->second * 1000L;
}


/* ------ Screen management ------ */

/* Initialize the console. */
/* do nothing, we did it in gp_init()! */
void
gp_init_console(P0())
{
}


/* Write a string to the console. */

void
gp_console_puts (const char *str, uint size)

{
	//fwrite (str, 1, size, stdout);
	zcPutbuf(theConsole, str, size);
}

/* Make the console current on the screen. */

int
gp_make_console_current (gx_device *dev)

{
	WindowPtr		w;
	
	w = zcWindow(theConsole);
	SelectWindow(w);
	
	return 0;
}

/* Make the graphics current on the screen. */

int
gp_make_graphics_current (gx_device *dev)
{
	return 0;
}

const char *
gp_getenv_display(P0())
{
	return NULL;
}

/* ------ Printer accessing ------ */

/* These should NEVER be called. */

/* Open a connection to a printer.  A null file name means use the */
/* standard printer connected to the machine, if any. */
/* "|command" opens an output pipe. */
/* Return NULL if the connection could not be opened. */

FILE *
gp_open_printer (char *fname, int binary_mode)
{
	if (strlen(fname) == 1  &&  fname[0] == '-')
		return stdout;
	else if (strlen(fname) == 0)
		return gp_open_scratch_file(gp_scratch_file_name_prefix, fname, binary_mode ? "wb" : "w");
	else
		return fopen(fname, binary_mode ? "wb" : "b");
}

/* Close the connection to the printer. */

void
gp_close_printer (FILE *pfile, const char *fname)
{
	fclose(pfile);
}


/* ------ File name syntax ------ */

/* Define the character used for separating file names in a list. */
const char gp_file_name_list_separator = ',';

/* Define the default scratch file name prefix. */
const char gp_scratch_file_name_prefix[] = "tempgs_";

/* Define the name of the null output file. */
const char gp_null_file_name[] = "????";

/* Define the name that designates the current directory. */
extern const char gp_current_directory_name[] = ":";

/* Define whether case is insignificant in file names. */
/* OBSOLETE
const int gp_file_names_ignore_case = 1;
*/

/* Define the string to be concatenated with the file mode */
/* for opening files without end-of-line conversion. */
const char gp_fmode_binary_suffix[] = "b";

/* Define the file modes for binary reading or writing. */
const char gp_fmode_rb[] = "rb";
const char gp_fmode_wb[] = "wb";


/* Create and open a scratch file with a given name prefix. */
/* Write the actual file name at fname. */

	FILE *
gp_open_scratch_file (const char *prefix, char *fname, const char *mode)

{
	strcpy (fname, (char *) prefix);
	{
		char newName[50];


		tmpnam (newName);
		strcat (fname, newName);
	}

	return fopen (fname, mode);
}


/* Answer whether a file name contains a directory/device specification, */
/* i.e. is absolute (not directory- or device-relative). */

	int
gp_file_name_is_absolute (const char *fname, register uint len)

{
	if (len /* > 0 */)
	{
		if (*fname == ':')
		{
			return 0;
		}
		else
		{
			register char  *p;
			register char	lastWasColon;
			register char	sawColon;


			for (len, p = (char *) fname, lastWasColon = 0, sawColon = 0;
				 len /* > 0 */;
				 len--, p++)
			{
				if (*p == ':')
				{
					sawColon = 1;

					if (lastWasColon /* != 0 */)
					{
						return 0;
					}
					else
					{
						lastWasColon = 1;
					}
				}
				else
				{
					lastWasColon = 0;
				}
			}

			return sawColon;
		}
	}
	else
	{
		return 0;
	}
}

/* Answer the string to be used for combining a directory/device prefix */
/* with a base file name.  The file name is known to not be absolute. */

const char *
gp_file_name_concat_string (const char *prefix, uint plen, const char *fname, uint len)

{
	if ( plen > 0 && prefix[plen - 1] == ':' )
		return "";
	return ":";
}


/* ------ File operations ------ */

/* If the file given by fname exists, fill in its status and return 1; */
/* otherwise return 0. */
/* OBSOLETE
	int
gp_file_status (const char *fname, file_status *pstatus)

{
#if 0
	struct stat sbuf;


	pstatus->size_pages		 = (sbuf.st_size + 511L) >> 9;
	pstatus->size_bytes		 = sbuf.st_size;
	pstatus->time_referenced = sbuf.st_mtime;
	pstatus->time_created	 = sbuf.st_ctime;
#endif

	return 1;
}
*/

/* ------ File enumeration ------ */

/****** THIS IS NOT SUPPORTED ON MACINTOSH SYSTEMS. ******/

struct file_enum_s {
	char *pattern;
	int first_time;
	gs_memory_t *memory;
};

/* Initialize an enumeration.  NEEDS WORK ON HANDLING * ? \. */

file_enum *
gp_enumerate_files_init (const char *pat, uint patlen, gs_memory_t *memory)

{	file_enum *pfen = 
		(file_enum *)gs_alloc_bytes(memory, sizeof(file_enum), "gp_enumerate_files");
	char *pattern;
	if ( pfen == 0 ) return 0;
	pattern = 
		(char *)gs_alloc_bytes(memory, patlen + 1, "gp_enumerate_files(pattern)");
	if ( pattern == 0 ) return 0;
	memcpy(pattern, pat, patlen);
	pattern[patlen] = 0;
	pfen->pattern = pattern;
	pfen->memory = memory;
	pfen->first_time = 1;
	return pfen;
}

/* Enumerate the next file. */

uint
gp_enumerate_files_next (file_enum *pfen, char *ptr, uint maxlen)

{	if ( pfen->first_time )
	   {	pfen->first_time = 0;
	   }
	return -1;
}

/* Clean up the file enumeration. */

void
gp_enumerate_files_close (file_enum *pfen)

{	
	gs_free_object(pfen->memory, pfen->pattern, "gp_enumerate_files_close(pattern)");
	gs_free_object(pfen->memory, (char *)pfen, "gp_enumerate_files_close");
}



int stat(const char * file_name, struct stat *buf)
{
	OSErr		err;
	HFileParam	hInfo;
	Str255		fName;
	
	strncpy((char *)&fName[1], file_name, 255);
	fName[0] = strlen(file_name);
	
	hInfo.ioVRefNum = 0;
	hInfo.ioFVersNum = 0;
	hInfo.ioFDirIndex = 0;
	hInfo.ioNamePtr = fName;
	hInfo.ioDirID = 0;
	
	err = PBHGetFInfoSync((HParmBlkPtr)&hInfo);
	if (err) return err;			// set errno?
	
	buf->st_blocks = (hInfo.ioFlPyLen + 511) >> 9;
	buf->st_size = hInfo.ioFlLgLen;
	buf->st_mtime = hInfo.ioFlMdDat;	// +/- some offset?
	buf->st_ctime = hInfo.ioFlCrDat;	// +/- some offset?
	
	return 0;
}

/* OBSOLETE
int unlink(const char *pathname)
{
	return remove(pathname);
}
*/


int
gp_check_interrupts(void)
{
	macHouseKeeping();
	
	if (interruptRequest) {
		interruptRequest = false;
		return 1;
	}
	
	return 0;
}

