/* Copyright (C) 1989 - 1995 Aladdin Enterprises.  All rights reserved.
  
  This file is part of Aladdin Ghostscript.
  
  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
  or distributor accepts any responsibility for the consequences of using it,
  or for whether it serves any particular purpose or works at all, unless he
  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
  License (the "License") for full details.
  
  Every copy of Aladdin Ghostscript must include a copy of the License,
  normally in a plain ASCII text file named PUBLIC.  The License grants you
  the right to copy, modify and redistribute Aladdin Ghostscript, but only
  under certain conditions described in the License.  Among other things, the
  License requires that the copyright notice and this notice be preserved on
  all copies.
*/

#include "MacHeaders"
#include <GestaltEqu.h>
#include "memory_.h"
#include "string_.h"
#include "gx.h"
#include "gp.h"
#include "gpcheck.h"
#include "gp_mac.h"
#include "gdebug.h"
#include "sys/stat.h"
#include <stdarg.h>
#include <console.h>
#if __profile__
#include "Profiler.h"
#endif

/* for consoles */
#include "zarfcon.h"
#include "iref.h"
#include "stream.h"
#include "gxiodev.h"			/* must come after stream.h */
#include "files.h"

#define DO_STREAM_SPOOFING

short		gWidth, gHeight;
short		gScreenDepth;

zcConsole	theConsole;
Boolean		interruptRequest = false;


static int
checkGestaltAttr(OSType selector, int attr)
{
	long		response;
	OSErr		error;
	
	error = Gestalt(selector, &response);
	if (error != noErr) return 0;
	return (response & (1 << attr)) != 0;
}


static int
checkGestaltValue(OSType selector, int value)
{
	long		response;
	OSErr		error;
	
	error = Gestalt(selector, &response);
	if (error != noErr) return 0;
	return (response & 0xFFFFL) >= value;
}


static void
InitGlobals (void)
{
	if (!checkGestaltAttr(gestaltAliasMgrAttr,		gestaltAliasMgrPresent))	goto tooOld;
	if (!checkGestaltAttr(gestaltAppleEventsAttr,	gestaltAppleEventsPresent))	goto tooOld;
	if (!checkGestaltAttr(gestaltFindFolderAttr,	gestaltFindFolderPresent))	goto tooOld;
	if (!checkGestaltAttr(gestaltFSAttr,			gestaltHasFSSpecCalls))		goto tooOld;
	if (!checkGestaltAttr(gestaltStandardFileAttr,	gestaltStandardFile58))		goto tooOld;
	if (!checkGestaltAttr(gestaltQuickdrawFeatures, gestaltHasColor))			goto tooOld;
	if (!checkGestaltAttr(gestaltQuickdrawFeatures, gestaltHasDeepGWorlds))		goto tooOld;
	if (!checkGestaltValue(gestaltQuickdrawVersion, gestalt32BitQD))			goto tooOld;
	{
		GDHandle		hGD	   = GetMainDevice();
		PixMapHandle	gdPMap = (*hGD)->gdPMap;

		gWidth 	= (*hGD)->gdRect.right  - (*hGD)->gdRect.left;
		gHeight	= (*hGD)->gdRect.bottom - (*hGD)->gdRect.top;

		gScreenDepth = (*gdPMap)->pixelSize;
	}
	return;

tooOld:
	Alert(301, nil);
	ExitToShell();
}


#define RESPONSE_INTERVAL		10

void
macHouseKeeping(void)
{
	EventRecord	e;
	Boolean didOne;
	static long lastTicks = 0;
	
	if ((unsigned long)(TickCount() - lastTicks) > RESPONSE_INTERVAL) {
		didOne = false;
		
		while (GetNextEvent(everyEvent, &e)) {
			ViewAppEvent(&e);
			didOne = true;
		}
		
		if (!didOne) {
			ViewAppEvent(&e);
		}
		
		lastTicks = TickCount();
	}
}



private void
create_console(void)
{
	zcCreateSpec		cSpec;
	WindowPtr			win;
	Rect				winRect;
	zcInit(macHouseKeeping);
	
	zcDefaultSpec(&cSpec);
	
	if (thePrefs.save_positions
	&&  thePrefs.console_size.v > 0
	&&  thePrefs.console_size.h > 0) {
		cSpec.rows = thePrefs.console_size.v;
		cSpec.cols = thePrefs.console_size.h;
	}
	else {
		cSpec.rows = 25;
		cSpec.cols = 80;
	}
	
	cSpec.font = monaco;
	cSpec.size = 9;
	cSpec.title = "Talk to Ghostscript";
	cSpec.delay = 0;
	cSpec.growable = 1;
	cSpec.visible = 0;
	cSpec.closeable = 1;
	
	theConsole = zcCreate(&cSpec);
	
	win = zcWindow(theConsole);
	
	if (thePrefs.save_positions
	&&  thePrefs.console_position.h != 0
	&&  thePrefs.console_position.v != 0) {
		MoveWindow(win,
			thePrefs.console_position.h,
			thePrefs.console_position.v,
			false);
	}
	else {
		winRect = win->portRect;
		MoveWindow(win,
			gWidth - (winRect.right - winRect.left) - 6,
			LMGetMBarHeight() + 26,
			false);
	}
	
	if (thePrefs.console_visible) {
		ShowWindow(win);
		SelectWindow(win);
	}
}

private void
destroy_console(void)
{
	zcDestroy(theConsole);
}


#ifdef DO_STREAM_SPOOFING

/* ====== Substitute for stdio ====== */
/* Forward references */
private void mac_std_init(void);
private stream_proc_process(mac_std_read_process);
private stream_proc_process(mac_std_write_process);

private int
mac_std_read_process(stream_state *st, stream_cursor_read *ignore_pr,
  stream_cursor_write *pw, bool last)
{	
	int count = pw->limit - pw->ptr;

	if ( count < 0 )		/* empty buffer */
		return 1;

	gp_check_interrupts();
	
	count = zcGetbuf(theConsole, (char *)(pw->ptr + 1), count);
	if (count <= 0) return 0;
	
	pw->ptr += count;	
	return 1;
}

private int
mac_std_write_process(stream_state *st, stream_cursor_read *pr,
  stream_cursor_write *ignore_pw, bool last)
{	
	uint count = pr->limit - pr->ptr;

	if ( count < 0 )		/* empty buffer */
		return 1;

	zcPutbuf(theConsole, (char *)(pr->ptr + 1), count); 
	pr->ptr += count;
	
	gp_check_interrupts();

	return 0;
}




/* Define alternate 'open' routines for our stdin/out/err streams. */

extern int iodev_stdin_open(P4(gx_io_device *, const char *, stream **,
			       gs_memory_t *));
private int
mac_stdin_open(gx_io_device *iodev, const char *access, stream **ps,
  gs_memory_t *mem)
{	int code = iodev_stdin_open(iodev, access, ps, mem);
	stream *s = *ps;
	if ( code != 1 )
	  return code;
	s->procs.process = mac_std_read_process;
	s->file = NULL;
	return 0;
}

extern int iodev_stdout_open(P4(gx_io_device *, const char *, stream **,
				gs_memory_t *));
private int
mac_stdout_open(gx_io_device *iodev, const char *access, stream **ps,
  gs_memory_t *mem)
{	int code = iodev_stdout_open(iodev, access, ps, mem);
	stream *s = *ps;
	if ( code != 1 )
	  return code;
	s->procs.process = mac_std_write_process;
	s->file = NULL;
	return 0;
}

extern int iodev_stderr_open(P4(gx_io_device *, const char *, stream **,
				gs_memory_t *));
private int
mac_stderr_open(gx_io_device *iodev, const char *access, stream **ps,
  gs_memory_t *mem)
{	int code = iodev_stderr_open(iodev, access, ps, mem);
	stream *s = *ps;
	if ( code != 1 )
	  return code;
	s->procs.process = mac_std_write_process;
	s->file = NULL;
	return 0;
}



/* Do one-time initialization */
/* reinitialize stdin/out/err to use our windows */
/* assume stream has already been initialized for the real stdin */
private iodev_proc_init(mac_stdio_init);
private int
mac_stdio_init(gx_io_device *iodev, gs_memory_t *mem)
{
	gs_findiodevice((byte *)"%stdin", 6)->procs.open_device = mac_stdin_open;
	gs_findiodevice((byte *)"%stdout", 7)->procs.open_device = mac_stdout_open;
	gs_findiodevice((byte *)"%stderr", 7)->procs.open_device = mac_stderr_open;

	return 0;
}



/* Use a pseudo IODevice to get mac_stdio_init called at the right time. */
/* This is bad architecture; we'll fix it later. */
gx_io_device gs_iodev_macstdio = {
	"macstdio", "Special",
	{ mac_stdio_init, iodev_no_open_device,
	  iodev_no_open_file, iodev_no_fopen, iodev_no_fclose,
	  iodev_no_delete_file, iodev_no_rename_file,
	  iodev_no_file_status, iodev_no_enumerate_files
	}
};



/* These are used instead of the stdio version. */
/* The declarations must be identical to that in <stdio.h>. */
int
fprintf(FILE *file, const char *fmt, ...)
{
	int		count;
	va_list	args;
	char	buf[1024];
	
	va_start(args,fmt);
	
	if (file != stdout  &&  file != stderr) {
		count = vfprintf(file, fmt, args);
	}
	else {
		count = vsprintf(buf, fmt, args);
		zcPutbuf(theConsole, buf, count);
	}
	
	va_end(args);
	return count;
}

int
fputs(const char *string, FILE *file)
{
	if (file != stdout  &&  file != stderr) {
		return fwrite(string, strlen(string), 1, file);
	}
	else {
		return (zcPuts(theConsole, string) == zcOK) ? 1 : EOF;
	}
}

#endif


short InstallConsole(short fd)
{
	if (theConsole == NULL)
		create_console();
	return 0;
}

void RemoveConsole(void)
{
	if (theConsole != NULL)
		destroy_console();
}

long WriteCharsToConsole(char *buffer, long n)
{
	int		err;
	
	err = zcPutbuf(theConsole, buffer, n);
	if (err == zcOK) err = zcPutc(theConsole, '\n');
	return (err == zcOK) ? n : -1;
}

long ReadCharsFromConsole(char *buffer, long n)
{
	long	count;
	
	macHouseKeeping();
	count = zcGetbuf(theConsole, buffer, n);
	return (count < 0) ? -1 : count;
}



/* this is the REAL main function */
#undef main
int gs_main(int argc, const char *argv[]);

#define MAX_ARGS		40

static void parseArgs(int * argc, char *argv[], char * input)
{
	static const char delimiters[] = " \t\n\r";
	
	for (; *argc < MAX_ARGS  &&  *input; (*argc)++) {
		// skip initial white space
		while (*input  &&  strchr(delimiters, *input) != NULL) input++;
		if (! *input) break;
		
		// point to next argument
		argv[*argc] = input;
		while (*input  &&  strchr(delimiters, *input) == NULL) input++;
		if (*input) *input++ = '\0';
	}
}


void
initAll(void)
{
	MaxApplZone();
	
	InitGraf(&qd.thePort);
	InitFonts();
	InitWindows();
	InitMenus();
	InitCursor();
	
	InitGlobals();

	interruptRequest = false;
	
	loadPreferences();	// have to load before creating console
	create_console();
	ViewAppInit();
#if __profile__
	ProfilerInit(collectDetailed, bestTimeBase, 100, 20);
#endif
}

void
exitAll(void)
{
#if __profile__
	ProfilerDump("\pMac GS Viewer.prof");
	ProfilerTerm();
#endif
	ViewAppExit();
	savePreferences(true);	// have to save before closing console
	destroy_console();
}


void
noMemoryExit(void)
{
	Alert(307, NULL);
	ExitToShell();
}


//#define COMMAND_LINE	"-"
#ifdef COMMAND_LINE
#define ADDITIONAL	"additional "
#else
#define ADDITIONAL
#endif

void main(void)
{
	char		* argv[MAX_ARGS + 1];
	int			argc;

#ifdef COMMAND_LINE
	char	fixedInput[] = COMMAND_LINE;
#endif

	char	userInput[256];
	
	initAll();
	
	argc = 0;
	parseArgs(&argc, argv, "Ghostscript");

#ifdef COMMAND_LINE
	parseArgs(&argc, argv, fixedInput);
	zcPuts(theConsole, "Compiled command line arguments: ");
	zcPuts(theConsole, fixedInput);
	zcPutc(theConsole, '\n');
#endif

	if (thePrefs.command_ask) {
		note_interp_state(INTERP_CMD_LINE);
		
		zcPuts(theConsole, "Enter " ADDITIONAL "command line arguments: ");
		while (zcGets(theConsole, userInput, 256) == zcNoInput) {
			EventRecord	e;
			
			WaitNextEvent(everyEvent, &e, 10, NULL);
			ViewAppEvent(&e);
		}
		zcPutc(theConsole, '\n');
		
		note_interp_state(INTERP_STARTUP);

		thePrefs.command_line[0] = strlen(userInput);
		if (thePrefs.command_line[0] > 0
		&&  userInput[thePrefs.command_line[0] - 1] == '\n')
			thePrefs.command_line[0] -= 1;
		memmove(&thePrefs.command_line[1], userInput,
			thePrefs.command_line[0]);
	}
	else {
		memmove(userInput, &thePrefs.command_line[1],
			thePrefs.command_line[0]);
		userInput[thePrefs.command_line[0]] = '\0';
		
		if (thePrefs.command_line[0] > 0) {
			zcPuts(theConsole, "Command line from Preferences: ");
			zcPuts(theConsole, userInput);
			zcPuts(theConsole, "\n\n");
		}
	}

	parseArgs(&argc, argv, userInput);
	argv[argc] = NULL;
	
	gs_main(argc, argv);
	
	exitAll();
		/* this never gets called from here as gs_main
		   calls gp_exit() and then exit().  Hence the
		   real call to exitAll() is in gp_exit(); */ 
}

