/* Copyright (C) 1994, 1995 Aladdin Enterprises.  All rights reserved.
  
  This file is part of Aladdin Ghostscript.
  
  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
  or distributor accepts any responsibility for the consequences of using it,
  or for whether it serves any particular purpose or works at all, unless he
  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
  License (the "License") for full details.
  
  Every copy of Aladdin Ghostscript must include a copy of the License,
  normally in a plain ASCII text file named PUBLIC.  The License grants you
  the right to copy, modify and redistribute Aladdin Ghostscript, but only
  under certain conditions described in the License.  Among other things, the
  License requires that the copyright notice and this notice be preserved on
  all copies.
*/

#include "MacHeaders"
#include "Palettes.h"
#include "Aliases.h"
#include "QDOffscreen.h"
#include "AppleEvents.h"
#include "Sound.h"
#include <stdio.h>
#include <stddef.h>
#include <string.h>
#include "gdevmac.h"
#include "gp_mac.h"
#include "ghost.h"
#include "main.h"
#include "errors.h"

extern gx_device_mac gs_mac_device;

static int					quitRequested;
static short				currDevItem = 1;
static ControlActionUPP		ourScrollProc;

//#define DO_DITHER
#ifdef DO_DITHER
#define COPY_MODE			(srcCopy | ditherCopy)
#else
#define COPY_MODE			(srcCopy)
#endif

static int					interpState = INTERP_STARTUP;

//#define TIME_TEST
#ifdef TIME_TEST
static unsigned long		startTime;
#endif

#define NUM_WAIT_CURSORS		10
#define WAIT_CURSOR_INTERVAL	10
#define WAIT_CURSOR_BASE		3000
static CursHandle			waitCursors[NUM_WAIT_CURSORS];

#define NUM_PAGE_CURSORS		2
#define PAGE_CURSOR_INTERVAL	30
#define PAGE_CURSOR_BASE		2000
static CursHandle			pageCursors[NUM_PAGE_CURSORS];

enum {
	menuApple		= 1000,
	menuFile		= 1001,
	menuEdit		= 1002,
	menuGhostscript	= 2000,
	menuDevices		= 2001,
	
	cmdAbout		= 100,
	cmdDeskAcc		= 101,
	
	cmdNew			= 200,
	cmdOpen			= 201,
	cmdClose		= 202,
	cmdSaveAs		= 204,
	cmdQuit			= 299,
	
	cmdPageSetUp	= 300,
	cmdPrint		= 301,
	
	cmdUndo			= 68,	// same as accUndo
	cmdCut			= 70,	// same as accCut
	cmdCopy			= 71,	// same as accCopy
	cmdPaste		= 72,	// same as accPaste
	cmdClear		= 73,	// same as accClear
	cmdSelectAll	= 75,

	cmdPreferences	= 400,
	cmdSettings		= 401,
	
	cmdCommandInput	= 500,
	cmdCommandOutput= 501,
	cmdNewOutput	= 502,
	cmdRun			= 510,
	cmdInterrupt	= 511,
	cmdNextPage		= 512,
	
	cmdChooseDevice	= 600,
	
	cmdUnknown		= -1,
	cmdNone			= 0
} ;

// this is the format of the MMap resource that maps menus to commands
typedef struct {
	short		menuID;
	short		menuItem;	// 0 means whole menu, -1 means remainder of menu
	short		command;
} MenuMapEntry;

static MenuMapEntry   **menuMap;
static int				menuMapSize;


static void
init_cursors(void)
{
	int			i;
	
	for (i = 0; i < NUM_WAIT_CURSORS; i++) {
		waitCursors[i] = GetCursor(WAIT_CURSOR_BASE + i);
	}
	
	for (i = 0; i < NUM_PAGE_CURSORS; i++) {
		pageCursors[i] = GetCursor(PAGE_CURSOR_BASE + i);
	}
}

static void
show_cursor(void)
{
	int				i;
	CursPtr			c;
	static int		lastC = -1;
	int				thisC;
	
	switch (interpState) {
		case INTERP_RUN:
		case INTERP_STARTUP:
			i = (TickCount() / WAIT_CURSOR_INTERVAL) % NUM_WAIT_CURSORS;
			c = *waitCursors[i];
			thisC = WAIT_CURSOR_BASE + i;
			break;
		
		case INTERP_PAUSE:
			i = (TickCount() / PAGE_CURSOR_INTERVAL) % NUM_PAGE_CURSORS;
			c = *pageCursors[i];
			thisC = PAGE_CURSOR_BASE + i;
			break;

		default:
		case INTERP_PROMPT:
		case INTERP_CMD_LINE:
			c = &qd.arrow;
			thisC = 0;
			break;
	}
	
	if (thisC != lastC) {
		SetCursor(c);
		lastC = thisC;
	}
}


static void
init_menu_map(void)
{
	Handle		mmapRes;
	
	mmapRes = GetResource('MMap', 128);
	menuMapSize = SizeResource(mmapRes) / sizeof(MenuMapEntry);
	menuMap = (MenuMapEntry **)mmapRes;
}

static short
map_menu_item(short menuID, short menuItem)
{
	int			i;
	
	for (i = 0; i < menuMapSize; i++)
		if ((*menuMap)[i].menuID == menuID
		&&  (*menuMap)[i].menuItem == menuItem)
			return (*menuMap)[i].command;
	
	if (menuItem != -1)
		// try for remainder menu entries...
		return map_menu_item(menuID, -1);
	
	return cmdUnknown;
}



static void
cmd_able(short command, int b)
{
	int			i, mi;
	
	for (i = 0; i < menuMapSize; i++)
		if ((*menuMap)[i].command == command) {
			mi = (*menuMap)[i].menuItem;
			if (mi >= 0) {
				if (b)
					EnableItem(GetMenu((*menuMap)[i].menuID), mi);
				else
					DisableItem(GetMenu((*menuMap)[i].menuID), mi);
				
				if (mi == 0)
					DrawMenuBar();
			}
			
			return;
		}
}


#define cmd_enable(c)	cmd_able(c, true)
#define cmd_disable(c)	cmd_able(c, false)



static void
map_cmd(short command, short * menuID, short * menuItem)
{
	int			i;
	
	for (i = 0; i < menuMapSize; i++)
		if ((*menuMap)[i].command == command) {
			*menuID = (*menuMap)[i].menuID;
			*menuItem = (*menuMap)[i].menuItem;
			return;
		}
	
	*menuID = 0;
	*menuItem = 0;
}


static void
convertSpecToPath(FSSpec * s, char * p, int pLen)
{
	DirInfo		block;
	Str255		dirName;
	int			totLen, dirLen;

	totLen = s->name[0];
	memcpy(p, s->name + 1, totLen);
	
	block.ioDrParID = s->parID;
	block.ioNamePtr = dirName;
	while (block.ioDrParID != fsRtParID) {
		block.ioVRefNum = s->vRefNum;
		block.ioFDirIndex = -1;
		block.ioDrDirID = block.ioDrParID;
		PBGetCatInfoSync((CInfoPBPtr)&block);
		
		dirLen = dirName[0];
		if (dirLen + totLen + 1 >= pLen - 1) goto error;
		memmove(p + dirLen + 1, p, totLen);
		memcpy(p, dirName + 1, dirLen);
		p[dirLen] = ':';
		totLen += dirLen + 1;
	}
	
	p[totLen] = 0;
	return;

error:
	p[0] = 0;
}



void
note_interp_state(int s)
{
	int			prevState;

	prevState = interpState;
	interpState = s;

	cmd_able(menuDevices, s == INTERP_PROMPT);

	if (s == INTERP_PROMPT) {
 		mac_gs_win	* mwin;
		mwin = macw_find_win(FrontWindow());
			
 		if (mwin != NULL
 		&&  mwin->mdev != NULL
		&&  ! mwin->initiallyVisible
 		) {
 			macw_detach_window(mwin);
#ifdef TIME_TEST
			{
				char		message[100];
				
				startTime = TickCount() - startTime;
				sprintf(message, "Took %ld ticks\n", startTime);
				zcPuts(theConsole, message);
			}
#endif
		}
 	}
 	show_cursor();
}

void
confirm_next_page(mac_gs_win * mwin)
{
	note_interp_state(INTERP_PAUSE);
	macw_close_picture(mwin);
	cmd_able(cmdNextPage, true);
	
	if (thePrefs.next_page_window)
		confirmDialogOpen();
	if (thePrefs.next_page_sound)
		SndPlay(NULL, (SndListHandle)GetNamedResource('snd ', "\pPageBell"), true);

	while (interpState == INTERP_PAUSE) {
		EventRecord		e;
		
		show_cursor();
		WaitNextEvent(everyEvent, &e, 10, NULL);
		ViewAppEvent(&e);
	}

	if (interruptRequest)
		macw_detach_window(mwin);

	cmd_able(cmdNextPage, false);
}


static void
set_menus(void)
{
	int			inPrompt;
	int			topWinGraphic;
	mac_gs_win	* mwin;
	
	mwin = macw_find_win(FrontWindow());
	
	inPrompt = interpState == INTERP_PROMPT;
	topWinGraphic = (mwin != NULL)
		&& ((mwin->mdev == NULL) || mwin->initiallyVisible);
	
	cmd_able(cmdNew,			inPrompt);
	cmd_able(cmdOpen,			interpState <= INTERP_PROMPT);
	cmd_able(cmdClose,			topWinGraphic);
	cmd_able(cmdSaveAs,			topWinGraphic);
//	cmd_able(cmdPrint,			topWinGraphic);
 
 	cmd_able(cmdCopy,			topWinGraphic);
 	
 	cmd_able(cmdCommandOutput,	currWindow != NULL);
 	cmd_able(cmdNewOutput,		inPrompt);
 	cmd_able(cmdRun,			inPrompt);
 	cmd_able(cmdInterrupt,		! inPrompt  &&  ! interruptRequest); 	
}



static void
runString(const char *str)
{
	if (interpState == INTERP_CMD_LINE) {
		zcTypec(theConsole, '\n');
	}
	
	zcTypes(theConsole, str);
}


static void
newWindow(char * title, Boolean visible, Boolean nl)
{
	char		message[200];
	
	if (currDevItem != 1)
		return;

	sprintf(message, "[%.1f %.1f] [%.1f %.1f] %d (%s) %s .newwindow%c",
		nextSpec.pixels_per_inch, nextSpec.pixels_per_inch,
		nextSpec.width, nextSpec.height,
		
		nextSpec.depth == SPEC_DEPTH_MONO ? 1
		: nextSpec.depth == SPEC_DEPTH_4BIT_GRAY ? 4
		: 8,
		
		title, visible ? "true" : "false",
		
		nl ? '\n' : ' ');

	runString(message);
}


static void
runFile(FSSpec * f, int doSave, int doNewWindow)
{
	char				fName[500];
	char				command[550];
	char				wName[100];

	convertSpecToPath(f, fName, sizeof(fName) - 1);
	memcpy(wName, &f->name[1], f->name[0]);
	wName[f->name[0]] = '\0';
	
	if (doSave)			runString("{ ");
	if (doNewWindow)	newWindow(wName, false, false);

	sprintf(command, "(%s) run", fName);
	runString(command);
	
	if (doSave)			runString(" } .runjob ");
	runString("\n");
}

static void
renderFile(FSSpec * f, FSSpec * fout, int doSave, int doNewWindow)
{
	char				fName[500];
	char				fOutName[500];
	char				command[550];

	if (doSave)			runString("{ ");

	convertSpecToPath(f, fName, sizeof(fName) - 1);
	
	if (doNewWindow) {
		convertSpecToPath(fout, fOutName, sizeof(fOutName) - 1);
		
		sprintf(command, "[%.1f %.1f] [%.1f %.1f] (%s) .newfile ",
			nextSpec.pixels_per_inch, nextSpec.pixels_per_inch,
			nextSpec.width, nextSpec.height,
			fOutName);
	
		runString(command);
	}

	sprintf(command, "(%s) run", fName);
	runString(command);
	
	if (doNewWindow) {
		runString(" .closefile");
	}
	
	if (doSave)			runString(" } .runjob ");
	runString("\n");
}

static void
chooseAndRunFile(int doSave, int doNewWindow)
{
	StandardFileReply	reply;
	OSType				osType[2];
	
	osType[0] = 'TEXT';
	osType[1] = 'EPSF';
	
	StandardGetFile(NULL, 2, osType, &reply);
	if (! reply.sfGood) return;
	
	{
		Boolean		targetIsFolder, wasAliased;
		
		ResolveAliasFile(&reply.sfFile, true,
			&targetIsFolder, &wasAliased);
		if (targetIsFolder) return;
	}
	
	if (currDevItem == 1) {
		runFile(&reply.sfFile, doSave, doNewWindow);
	}
	else {
		StandardFileReply	replyOut;
		Str255				oName;
		Str255				dName;
		
		if (doNewWindow) {
			short		deviceMenuID, deviceItem;
			MenuHandle	deviceMenu;
		
			map_cmd(menuDevices, &deviceMenuID, &deviceItem);
			deviceMenu = GetMHandle(deviceMenuID);
			GetItem(deviceMenu, currDevItem, dName);
			
			memcpy(&oName[1], &reply.sfFile.name[1], reply.sfFile.name[0]);
			oName[reply.sfFile.name[0]+1] = '.';
			memcpy(&oName[reply.sfFile.name[0]+2], &dName[1], dName[0]);
			oName[0] = reply.sfFile.name[0] + 1 + dName[0];
		
			StandardPutFile("\pSave rendered image as:", oName, &replyOut);
			if (! replyOut.sfGood) return;
		}
		
		renderFile(&reply.sfFile, &replyOut.sfFile, doSave, doNewWindow);
	} 

#ifdef TIME_TEST
	startTime = TickCount();
#endif
}


static pascal void
doScrollProc (ControlHandle	whichCtl, short iCtlPart)
{
	mac_gs_win	*mwin;
	Boolean		vertical;
	short		curr, min, max;
	short		next;
		
	mwin = (mac_gs_win *)GetCRefCon(whichCtl);
	vertical = whichCtl == mwin->vScrollBar;
	
	curr = vertical ? mwin->win.origin.v : mwin->win.origin.h;
	
	switch (iCtlPart)
	{
		case 0:
			return;
			
		case inUpButton:
			next = curr - 5;
			break;

		case inDownButton:
			next = curr + 5;
			break;

		case inThumb:
			next = GetCtlValue(whichCtl);
			break;

		case inPageUp:
			if (vertical)
				next = curr - (mwin->win.viewRect.bottom - mwin->win.viewRect.top);
			else
				next = curr - (mwin->win.viewRect.right - mwin->win.viewRect.left);
			break;

		case inPageDown:
			if (vertical)
				next = curr + (mwin->win.viewRect.bottom - mwin->win.viewRect.top);
			else
				next = curr + (mwin->win.viewRect.right - mwin->win.viewRect.left);
			break;
	}

	min = GetCtlMin(whichCtl);
	max = GetCtlMax(whichCtl);
	if (min > next)			next = min;
	else if (max < next)	next = max;

	if (next != curr) {
		GrafPtr				savePort;
		Rect				rPict, rWin;
		short				dx, dy;

		if (vertical)		dx = 0, dy = curr - next;
		else				dx = curr - next, dy = 0;
		
		if (iCtlPart != inThumb)
			SetCtlValue(whichCtl, next);

		GetPort(&savePort);
		SetPort((GrafPtr)mwin->win.port);

		/*...Scroll the previous update region...*/
		OffsetRgn(((WindowPeek)mwin->win.port)->updateRgn, dx, dy);

		/*...Offset the origin...*/
		mwin->win.origin.h -= dx;
		mwin->win.origin.v -= dy;

		/*...Update the Display...*/
		if (!mwin->pixmapLocked) LockPixels(mwin->pixmap);
		ForeColor(blackColor);
		BackColor(whiteColor);
		
		rWin = mwin->win.viewRect;
		
		if ((*mwin->win.port->visRgn)->rgnSize == 10
		&&  PtInRect(topLeft(rWin), &(*mwin->win.port->visRgn)->rgnBBox) 
		&&  PtInRect(botRight(rWin), &(*mwin->win.port->visRgn)->rgnBBox)) {
			/* The viewRect is entirely visible, so use a quick scroll */
			
			/*...Scroll the image...*/
			ScrollRect(&(mwin->win.viewRect), dx, dy, NULL);
		
			if (dy > 0)			rWin.bottom = rWin.top + dy;
			else if (dy < 0)	rWin.top = rWin.bottom + dy;
			else if (dx > 0)	rWin.right = rWin.left + dx;
			else 				rWin.left = rWin.right + dx;
			rPict = rWin;
			OffsetRect(&rPict, mwin->win.origin.h, mwin->win.origin.v);
			
			CopyBits((BitMap *)*mwin->pixmap, &((GrafPtr)mwin->win.port)->portBits,
				 &rPict, &rWin, COPY_MODE, nil);
		}
		else {
			/* the viewRect is partially obscured, so just copy the whole thing */
zcPuts(theConsole, "ICK!\n");
			rPict = rWin;
			OffsetRect(&rPict, mwin->win.origin.h, mwin->win.origin.v);
		
			CopyBits((BitMap *)*mwin->pixmap, &((GrafPtr)mwin->win.port)->portBits,
				 &rPict, &rWin, COPY_MODE, nil);
		}
		
		if (!mwin->pixmapLocked) UnlockPixels(mwin->pixmap);

		SetPort(savePort);
	}
}


static void
doGWinUpdate(mac_gs_win * mwin, WindowPtr window)
{
	Rect			viewRect;
	
	SetPort((GrafPtr)window);
	
	BeginUpdate(window);

	viewRect = mwin->pict.viewRect;
	OffsetRect(&viewRect, -mwin->win.origin.h, -mwin->win.origin.v);

	ClipRect(&(mwin->win.viewRect));
	
	if (!mwin->pixmapLocked) LockPixels(mwin->pixmap);
	
	ForeColor(blackColor);
	BackColor(whiteColor);
	CopyBits((BitMap *)*mwin->pixmap, &window->portBits,
		 &mwin->pict.viewRect, &viewRect, COPY_MODE, nil);
		 
	if (!mwin->pixmapLocked) UnlockPixels(mwin->pixmap);
	
	ClipRect(&mwin->win.port->portRect);

	UpdateControls(window, window->visRgn);
	DrawGrowIcon(window);

	EndUpdate(window);
}



static void
doGWinActivate(mac_gs_win * mwin, WindowPtr window, int activate)
{
	HiliteControl(mwin->hScrollBar, activate ? 0 : 255);
	HiliteControl(mwin->vScrollBar, activate ? 0 : 255);

#ifdef WIN_PALETTE
	if (activate)
		ActivatePalette(mwin->win.port);
#endif
}



static void
doCommand(short cmd, short menuID, short menuItem)
{
	mac_gs_win	* mwin;
	
	mwin = macw_find_win(FrontWindow());
	
	switch (cmd) {
		case cmdPageSetUp:
		case cmdPrint:
			break;

		case cmdAbout: {
			aboutDialogOpen();
			break;
		}
			
		case cmdDeskAcc: {
			GrafPtr		savePort;
			Str255		name;
			
			GetPort(&savePort);
			GetItem(GetMHandle(menuID), menuItem, name);
			// should fix menus here...
			OpenDeskAcc(name);
			SetPort(savePort);
			break;
		}
		
		case cmdCut:			// by default cut is same as copy
		case cmdCopy:
			if (mwin != NULL) {
				macw_copy_picture(mwin);
			}
			break;

		case cmdPaste:
			break;

		case cmdClear:
			if (mwin != NULL) {
				macw_clear_window(mwin);
			}
			break;

		case cmdNew:
		case cmdNewOutput:
			newWindow("Rendering Output", true, true);
			break;

		case cmdOpen:
			chooseAndRunFile(true, true);
			break;

		case cmdClose:
			if (mwin != NULL
			&& (mwin->mdev == NULL  ||  interpState == INTERP_PROMPT)) {
				macw_detach_window(mwin);
				macw_close_window(mwin);
			}
			break;

		case cmdSaveAs:
			if (mwin != NULL) {
				macw_save_picture(mwin);
			}
			break;		
		
		case cmdQuit:
			if (interpState != INTERP_PROMPT) {
				interruptRequest = true;
				note_interp_state(INTERP_RUN);
			}
			quitRequested = true;
			runString("quit\n");
			break;

		case cmdCommandInput:
			ShowWindow(zcWindow(theConsole));
			SelectWindow(zcWindow(theConsole));
			break;
			
		case cmdCommandOutput:
			if (currWindow != NULL)
				macw_top_window(currWindow);
			break;
			
		case cmdRun:
			chooseAndRunFile(false, false);
			break;

		case cmdInterrupt:
			confirmDialogDone(true);		// in case we are in a confirm
			interruptRequest = true;
			note_interp_state(INTERP_RUN);
			break;
		
		case cmdNextPage:
			confirmDialogDone(false);		// in case confirm dialog is up
			note_interp_state(INTERP_RUN);
			break;
			
		case cmdChooseDevice: {
			MenuHandle		devMenu;
			Str255			devStr;
			char			command[300];
			
			devMenu = GetMenu(menuID);
			GetItem(devMenu, menuItem, devStr);
			CheckItem(devMenu, currDevItem, false);
			CheckItem(devMenu, menuItem, true);
			currDevItem = menuItem;
			sprintf(command, "(%.*s) selectdevice\n", devStr[0], devStr + 1);
			runString(command);
			break;
		}
		
		case cmdSettings:
			specDialogOpen();
			break;
			
		case cmdPreferences:
			prefDialogOpen();
			break;
	}
}


static void
doMenu(long menu)
{
	short		menuID, menuItem;
	short		cmd;
	
	menuID = HiWord(menu);
	menuItem = LoWord(menu);
	cmd = map_menu_item(menuID, menuItem);
	
	doCommand(cmd, menuID, menuItem);
	
	HiliteMenu(0);
}		


/*
 *	doKeyEvent -- handles keyboard events
 *
 */

static void
doKey(EventRecord *pEvent)
{
	char			aChar		= pEvent->message & charCodeMask;
	short			iCtlPart;
	mac_gs_win	*mwin;
	ControlHandle	hControl;

	if (pEvent->modifiers & cmdKey) {
		long		menuResult;

		set_menus();
		
		menuResult = MenuKey(aChar);
		
		if (HiWord(menuResult) == 0) {
			long	transKey;
			short	keyNoCmd;
			long	deadState;
			Ptr		theKCHR;
			
			keyNoCmd = (pEvent->message & keyCodeMask) >> 8
						| (pEvent->modifiers & 0xFE00);	// no command bit
			deadState = 0;
			theKCHR = (Ptr)GetEnvirons(smKCHRCache);
			
			transKey = KeyTrans(theKCHR, keyNoCmd, &deadState);
			aChar = (transKey & 0x00FF0000) >> 16;
			if (aChar == 0) aChar = (transKey & 0x000000FF);
			if (aChar == 0) return;
			
			menuResult = MenuKey(aChar);
		}
		
		doMenu(menuResult);
		return;
	}

	if ('a' <= aChar && aChar <= 'z')
		aChar += 'A' - 'a';

	if (interpState == INTERP_PAUSE
	&& (aChar == ' '  ||  aChar == '\r'  ||  aChar == '\003')) {
		doCommand(cmdNextPage, 0, 0);
		return;
	}
	
	mwin = macw_find_win(FrontWindow());
	if (mwin == NULL) return;
	
	switch (aChar)
	{
		case 0x1C:	// left arrow
		case 0x1D:	// right arrow
			hControl = mwin->hScrollBar;
			break;

		case 0x1E:	// up arrow
		case 0x1F:	// down arrow
			hControl = mwin->vScrollBar;
			break;

		default:
			hControl = (pEvent->modifiers & optionKey)
					   		? mwin->hScrollBar
					   		: mwin->vScrollBar;
	}

	switch (aChar)
	{
		case 0x1C:	// left arrow
		case 0x1E:	// up arrow
			iCtlPart = inUpButton;
			break;

		case 0x1D:	// right arrow
		case 0x1F:	// down arrow
			iCtlPart = inDownButton;
			break;

		case 0x0B:	// page up (& ^K)
			iCtlPart = inPageUp;
			break;

		case 0x0C:	// page down (& ^L)
			iCtlPart = inPageDown;
			break;

		case 0x01:	// home (& ^A)
			iCtlPart = inThumb;
			SetCtlValue (hControl, GetCtlMin (hControl));
			break;
			
		case 0x04:	// end (& ^D)
			iCtlPart = inThumb;
			SetCtlValue (hControl, GetCtlMax (hControl));
			break;
		
		default:
			return;
	}

	doScrollProc(hControl, iCtlPart);
}




static void
doMouseDown(EventRecord *pEvent)
{
	WindowPtr		window;
	mac_gs_win		* mwin;
	short			part;

	part = FindWindow(pEvent->where, &window);
	mwin = macw_find_win(window);
	
	switch (part)
	{
		case inDesk:
			break;
			
		case inMenuBar:
			set_menus();
			doMenu(MenuSelect(pEvent->where));
			break;

		case inSysWindow: 
			SystemClick(pEvent, window);
			break;
	    
		case inContent:
		{
			ControlHandle	whichCtl;
			short			iCtlPart;

			if (window != FrontWindow()) {
				SelectWindow(window);
				break;
			}
			
			if (mwin == NULL) break;
			SetPort((GrafPtr) window);
			GlobalToLocal (&pEvent->where);

			if ((iCtlPart = FindControl (pEvent->where, window, &whichCtl)) > 0)
			{
				iCtlPart = TrackControl(whichCtl, pEvent->where,
										 (iCtlPart != inThumb)
										 	 ? ourScrollProc
										 	 : NULL);
				if (iCtlPart == inThumb)
					doScrollProc(whichCtl, iCtlPart);
			}

			break;
		}

		case inDrag:
		{
			Rect	dragRect;
			
			dragRect = qd.screenBits.bounds;
			InsetRect(&dragRect, 4, 4);
			DragWindow(window, pEvent->where, &dragRect);
			break;
		}
		
		case inGoAway:
			if (TrackGoAway(window, pEvent->where))
				doCommand(cmdClose, 0, 0);
			break;

		case inGrow:
		{
			long newSize;
			Rect growRect;

			if (mwin == NULL) break;
			SetPort((GrafPtr) window);

			growRect.top	= 100;
			growRect.bottom = mwin->pict.viewRect.bottom + 16;
			growRect.left	= 200;
			growRect.right  = mwin->pict.viewRect.right + 16;
			
			newSize = GrowWindow(window, pEvent->where, &growRect);
			
			if (newSize != 0L) {
				macw_inval_window(mwin);
				SizeWindow(window, LoWord (newSize), HiWord (newSize), TRUE);
				macw_redraw_window(mwin);
			}
			
			break;
		}


		case inZoomIn:
		case inZoomOut:
		{
			Rect   *pRect = &(((WindowPeek) window)->port.portRect);

			if (mwin == NULL) break;
			SetPort((GrafPtr)window);

			if (TrackBox(window, pEvent->where, part)) {
				macw_inval_window(mwin);
				ZoomWindow(window, part, FALSE);
				macw_redraw_window(mwin);
			}
			
			break;
		}
	}
}



/*
 *	doEvent -- handles events
 *
 */

static void
doEvent (EventRecord *pEvent)
{
	WindowPtr		window;
	mac_gs_win		* mwin;
	

	if (zcHandleEvent(pEvent)) goto done;
	if (dialogHandleEvent(pEvent)) goto done;

	switch (pEvent->what) {
		case mouseDown:
			doMouseDown(pEvent);
			break;
		
		case autoKey:
		case keyDown:
			doKey(pEvent);
			break;

		case updateEvt:
			window = (WindowPtr)pEvent->message;
			mwin = macw_find_win(window);
			if (mwin != NULL) {
				doGWinUpdate(mwin, window);
			}
			else {
				// we don't know whose window this is,
				// but we don't want to keep getting the
				// events for it.
				BeginUpdate(window);
				EndUpdate(window);
			}
			break;

		case activateEvt:
			window = (WindowPtr)pEvent->message;
			mwin = macw_find_win(window);
			if (mwin != NULL)
				doGWinActivate(mwin, window, (pEvent->modifiers & activeFlag) != 0);
			break;
			
		case kHighLevelEvent:
			AEProcessAppleEvent(pEvent);
			break;
	}

done:
	;
}






static pascal OSErr
openAppHandler(AppleEvent *theAppleEvent, AppleEvent *reply, long handlerRefcon)
{
	return noErr;
}

static pascal OSErr
quitHandler(AppleEvent *theAppleEvent, AppleEvent *reply, long handlerRefcon)
{
	return noErr;
}

static pascal OSErr
openDocsHandler(AppleEvent *theAppleEvent, AppleEvent *reply, long handlerRefcon)
{
	OSErr		err;
	AEDescList	docList;
	long		index, itemsInList;
	FSSpec		fSpec;
	Size		actualSize;
	AEKeyword	keywd;
	DescType	returnedType;

	// get the direct parameter--a descriptor list--and put
	// it into docList
	err = AEGetParamDesc(theAppleEvent, keyDirectObject, typeAEList, &docList);
	if (err)	return err;

	// count the number of descriptor records in the list
	err = AECountItems(&docList, &itemsInList);
	if (err)	goto done;

	// now get each descriptor record from the list, coerce
	// the returned data to an FSSpec record, and open the
	// associated file
	for (index = 1; index <= itemsInList; index++) {
		err = AEGetNthPtr(&docList, index, typeFSS, &keywd, &returnedType,
			(Ptr)&fSpec, sizeof(fSpec), &actualSize);
		if (err) goto done;
		
		runFile(&fSpec, true, true);
		
		break;
	}
	

done:
	AEDisposeDesc(&docList);

	return	err;
}




/*
 *  ViewAppInit -- 
 *
 */

void
ViewAppInit(void)
{
	Handle		menuBar;
	
	init_cursors();
	init_menu_map();
	menuBar = GetNewMBar(128);
	SetMenuBar(menuBar);
	
	{
		short		deviceMenuID, deviceItem;
		Str255		item;
		MenuHandle	deviceMenu;
		
		map_cmd(menuDevices, &deviceMenuID, &deviceItem);
				
		if (deviceMenuID != 0  &&  deviceItem == 0) {
			deviceMenu = GetMHandle(deviceMenuID);
			
#define device_(d) {\
			extern gx_device	d;			\
			strcpy((char *)&(item[1]), d.dname);	\
			item[0] = strlen(d.dname);		\
			AppendMenu(deviceMenu, item);	\
			}
#include "gconfig.h"
#undef device_

			CheckItem(deviceMenu, 1, true);
		}
	}
	{
		short		appleMenuID, appleItem;
		MenuHandle	appleMenu;
		
		map_cmd(menuApple, &appleMenuID, &appleItem);
		appleMenu = GetMHandle(appleMenuID);
		AddResMenu(appleMenu, 'DRVR');
	}
	{
		AEEventHandlerUPP	openAppUPP, openDocsUPP, quitUPP;
		
		openAppUPP = NewAEEventHandlerProc(openAppHandler);
		openDocsUPP = NewAEEventHandlerProc(openDocsHandler);
		quitUPP = NewAEEventHandlerProc(quitHandler);
		
		AEInstallEventHandler(kCoreEventClass, kAEOpenApplication, openAppUPP, 0, false);
		AEInstallEventHandler(kCoreEventClass, kAEOpenDocuments, openDocsUPP, 0, false);
		AEInstallEventHandler(kCoreEventClass, kAEQuitApplication, quitUPP, 0, false);
	}
	
	note_interp_state(INTERP_STARTUP);
	// DrawMenuBar(); done by above
	
	
	ourScrollProc = NewControlActionProc(doScrollProc);
	quitRequested = false;
	
	if (thePrefs.specs_visible)
		specDialogOpen();
		
	// note_interp_state(INTERP_RUN); done by executive in gs_init.ps
}


/*
 *  ViewAppExit -- exits the viewer all
 *
 */

void
ViewAppExit(void)
{
	if (! quitRequested) {
		zcFlush(theConsole);
		zcPuts(theConsole, "\nGhostScript Terminated.\nType return to quit...");
	
		cmd_disable(menuDevices);
		cmd_disable(cmdNew);
		cmd_disable(cmdOpen);
		cmd_disable(cmdSettings);
		cmd_disable(cmdCommandOutput);
		cmd_disable(cmdNewOutput);
		cmd_disable(cmdRun);
		cmd_disable(cmdInterrupt);
		
		while (zcGetc(theConsole) == zcNoInput) {
			EventRecord			e;
			
			if (GetNextEvent(everyEvent, &e))
				doEvent(&e);
		}
	}
}


void
ViewAppEvent(EventRecord *pEvent)
{
	if (interpState != INTERP_PROMPT)
		show_cursor();
	
	doEvent(pEvent);
}


