/* zarfcon.c

	This is the implementation of the ZarfCon console package.
	
	Copyright (C) 1994 Glyphic Technology
	
	Permission to use, copy, modify, distribute, and sell this software and
	its documentation for any purpose is hereby granted without fee, provided
	that the above copyright notice and this permission notice appear in all
	copies of the software and related documentation.
	
	To Do:
		handle an optional close box & grow box
		handle optional non-auto-wrap
		handle tab settings
 */

#include "MacHeaders"
#include <string.h>
#include "zarfcon.h"

#define MAX_CONSOLES			5
#define INPUT_BUFFER_SIZE		1024

#define TOP_MARGIN				2
#define LEFT_MARGIN				2
#define BOTTOM_MARGIN			1
#define RIGHT_MARGIN			2

#define TREAT_CR_AS_NL			0

typedef struct {
	WindowPtr		window;
	int				charW, charH;
	int				charBase;
	int				curRow, curCol;
	int				numRows, numCols;
	
	int				echo:1;
	int				raw:1;
	int				delay:1;
	int				growable:1;
	
	int				cursorHideCount;		// for hiding during drawing
	int				cursorBlink:1;			// blinked on and off if at top
	int				cursorActive:1;			// if top window
	
	char			input[INPUT_BUFFER_SIZE];
	int				inStart;
	int				inLine;
	int				inEnd;
	
	char			** image;
	
	Rect			growRect;
	RgnHandle		clipRgn;
} zcState;
	
typedef struct {
	zcEventMgrProc	*mgr;
	unsigned long	lastBlink;
	zcState			**(consoles[MAX_CONSOLES]);
} zcGlobals;		

static zcGlobals		** g = NULL;



static zcConsole zc_find_win(WindowPtr w)
{
	int			i;
	zcState		** s;
	
	if (g == NULL) return zcNoConsole;

	for (i = 0; i < MAX_CONSOLES; i++) {
		s = (*g)->consoles[i];
		if (s != NULL && (*s)->window == w)
			return i;
	}
	
	return zcNoConsole;
}



static zcState * zc_use(zcConsole c)
{
	zcState		** sHandle;
	
	if (g == NULL) return NULL;
	if (c < 0  ||  c >= MAX_CONSOLES) return NULL;
	
	sHandle = (*g)->consoles[c];
	if (sHandle == NULL) return NULL;
	
	HLock((Handle)sHandle);
	return *sHandle;
}


static void zc_done(zcConsole c)
{
	zcState		** sHandle;
	
	if (g == NULL) return;
	if (c < 0  ||  c >= MAX_CONSOLES) return;
	
	sHandle = (*g)->consoles[c];
	if (sHandle == NULL) return;
	
	HUnlock((Handle)sHandle);
}


void zcInit(zcEventMgrProc *mgr)
{
	int		i;
	
	if (g != NULL) return;

	g = (zcGlobals **)(NewHandle(sizeof(zcGlobals)));
	
	(*g)->mgr = mgr;
	(*g)->lastBlink = 0;
	
	for (i = 0; i < MAX_CONSOLES; i++)
		(*g)->consoles[i] = NULL;
}


void zcExit()
{
	int		i;
	
	if (g == NULL) return;
	
	for (i = 0; i < MAX_CONSOLES; i++)
		if ((*g)->consoles[i] != NULL)
			zcDestroy(i);
	
	DisposeHandle((Handle)g);
	g = NULL;
}


#define SHOW		1
#define HIDE		-1
#define PAINT		0


static void zc_cursor(zcState * state, int msg)
{
	Rect		cr;
	int			oldState, newState;
	
	
	oldState = state->cursorHideCount == 0  &&  state->cursorBlink;
	
	switch (msg) {
		case PAINT:		goto doit;
		case SHOW:		state->cursorHideCount -= 1;	break;
		case HIDE:		state->cursorHideCount += 1;	break;
	}
	
	newState = state->cursorHideCount == 0  &&  state->cursorBlink;
	
	if (newState != oldState) {
doit:
		cr.top = state->curRow * state->charH + TOP_MARGIN;
		cr.left = state->curCol * state->charW + LEFT_MARGIN;
		cr.bottom = cr.top + state->charH;
		cr.right = cr.left + state->charW;
	
		if (state->cursorActive) {
			InvertRect(&cr);
		}
		else {
			PenMode(patXor);
			FrameRect(&cr);
		}
	}
}


static void zc_resize(zcState * state, int newRows, int newCols)
{
	Handle		oldImageH, newImageH;
	char		* oldImage, * newImage;
	GrafPtr		savePort;
	int			or, nr, nc, r;
	
	if (newRows == state->numRows
	&&  newCols == state->numCols)
		return;
		
	GetPort(&savePort);
	SetPort(state->window);
	
	zc_cursor(state, HIDE);

	newImageH = NewHandle((long)newRows * (long)newCols);
	if (newImageH == NULL) goto done;
	
	HLock(newImageH);
	newImage = *newImageH;
	
	memset(newImage, ' ', (long)newRows * (long)newCols);
	
	if (state->image != NULL) {
		oldImageH = state->image;
		HLock(oldImageH);
		oldImage = *oldImageH;
		
		if (newRows < state->curRow + 1)
			nr = newRows,			or = state->curRow - newRows + 1;
		else if (newRows < state->numRows)
			nr = newRows,			or = 0;
		else
			nr = state->numRows,	or = 0;
		
		if (newCols >= state->numCols)	nc = state->numCols;
		else							nc = newCols;
		
		for (r = 0; r < nr; r++) {
			memcpy(newImage + (r * newCols),
				oldImage + ((r + or) * state->numCols),
				nc);
		};
	
		DisposeHandle(oldImageH);
	}
	else {
		or = 0;
	}

	HUnlock(newImageH);
	state->image = (char **)newImageH;
	
	state->numRows = newRows;
	state->numCols = newCols;
	
	state->curRow -= or;
	if (state->curRow < 0) state->curRow = 0;
	if (state->curCol > newCols) state->curCol = newCols;
	
	SizeWindow(state->window,
		newCols * state->charW + LEFT_MARGIN + RIGHT_MARGIN,
		newRows * state->charH + TOP_MARGIN + BOTTOM_MARGIN,
		true);
	InvalRect(&state->window->portRect);

	RectRgn(state->clipRgn, &state->window->portRect);
	
	if (state->growable) {
		static RgnHandle	growRgn = NULL;
		
		state->growRect = state->window->portRect;
		state->growRect.top = state->growRect.bottom - 15;
		state->growRect.left = state->growRect.right - 15;

		if (growRgn == NULL)
			growRgn = NewRgn();
			
		RectRgn(growRgn, &state->growRect);
		DiffRgn(state->clipRgn, growRgn, state->clipRgn);
		//DisposeRgn(growRgn);
	}
	
	SetClip(state->clipRgn);
	
done:
	zc_cursor(state, SHOW);

	SetPort(savePort);
}


void zcDefaultSpec(zcCreateSpec *s)
{
	s->loc.h = 10;
	s->loc.v = 50;
	s->rows = 24;
	s->cols = 80;
	s->font = monaco;
	s->size = 9;
	s->title = "Console";
	s->echo = 1;
	s->raw = 0;
	s->delay = 1;
	s->visible = 1;
	s->growable = 0;
	s->closeable = 0;
}


zcConsole zcCreate(zcCreateSpec *s)
{
	int			i, l;
	int			cw, ch, cb;
	zcState		** sHandle, * state;
	Str255		ptitle;
	WindowPtr	w;
	GrafPtr		savePort;
	FontInfo	fontInfo;
	Rect		area;
	
	if (g == NULL) zcInit(zcHouseKeeping);
	
	for (i = 0; i < MAX_CONSOLES; i++)
		if ((*g)->consoles[i] == NULL)
			goto gotFree;
	
	return zcNoConsole;

gotFree:
	l = strlen(s->title);
	if (l > 255) l = 255;
	memmove(&ptitle[1], s->title, l);
	ptitle[0] = l;
	
	area.top = s->loc.v;
	area.left = s->loc.h;
	area.bottom = area.top + 10;
	area.right = area.left + 20;
	
	w = NewWindow(NULL, &area, ptitle,
		false, s->growable ? documentProc : noGrowDocProc,
		(WindowPtr)-1, s->closeable, 0);
		
	GetPort(&savePort);
	SetPort(w);
	TextMode(srcCopy);
	TextFont(s->font);
	TextSize(s->size);
	GetFontInfo(&fontInfo);
	ch = fontInfo.ascent + fontInfo.descent + fontInfo.leading;
	//cw = fontInfo.widMax;
	cw = CharWidth('X');
	cb = fontInfo.ascent;
	SetPort(savePort);
		
	sHandle = (zcState **)NewHandle(sizeof(zcState));
	HLock((Handle)sHandle);
	state = *sHandle;
	
	state->window = w;
	state->charW = cw;
	state->charH = ch;
	state->charBase = cb;
	//state->curRow = s->rows - 1;
	state->curRow = 0;
	state->curCol = 0;
	state->numRows = 0;	// set by call to zc_resize
	state->numCols = 0;
	state->inStart = 0;
	state->inLine = 0;
	state->inEnd = 0;
	state->image = NULL;
	state->echo = s->echo;
	state->raw = s->raw;
	state->delay = s->delay;
	state->growable = s->growable;
	state->cursorHideCount = 0;
	state->cursorBlink = 1;
	state->cursorActive = 0;
	state->clipRgn = NewRgn();

	zc_resize(state, s->rows, s->cols);
	
	HUnlock((Handle)sHandle);
	(*g)->consoles[i] = sHandle;

	if (s->visible)
		ShowWindow(w);

	return i;
}


void zcDestroy(zcConsole c)
{
	zcState		** state;
	
	if (g == NULL) return;
	if (c < 0  ||  c >= MAX_CONSOLES) return;
	
	state = (*g)->consoles[c];
	if (state == NULL) return;
	
	DisposeWindow((*state)->window);
	DisposeRgn((*state)->clipRgn);
	DisposeHandle((Handle)(*state)->image);
	DisposeHandle((Handle)state);
	
	(*g)->consoles[c] = NULL;
}


void zcGetSpec(zcConsole c, zcCreateSpec *s)
{
	zcState			* state;
	WindowPtr		w;
	Str255			ptitle;
	int				l;
	
	state = zc_use(c);
	if (state == NULL) {
		memset(s, 0, sizeof(zcCreateSpec));
		return;
	}
	
	w = state->window;

	s->loc.v = w->portRect.top - w->portBits.bounds.top;
	s->loc.h = w->portRect.left - w->portBits.bounds.left;
	s->rows = (w->portRect.bottom - w->portRect.top) / state->charH;
	s->cols = (w->portRect.right - w->portRect.left) / state->charW;
	s->font = w->txFace;
	s->size = w->txSize;
	
	if (s->title != NULL) {
		GetWTitle(w, ptitle);
		l = ptitle[0];
		memmove(s->title, &ptitle[1], l);
		s->title[l] = '\0';
	}
	
	s->echo = state->echo;
	s->raw = state->raw;
	s->delay = state->delay;
	s->visible = ((WindowPeek)w)->visible;
	s->growable = state->growable;

	zc_done(c);
}



WindowPtr zcWindow(zcConsole c)
{
	zcState		* state;
	WindowPtr	w;
	
	state = zc_use(c);
	if (state == NULL) return NULL;
	
	w = state->window;
	zc_done(c);
	return w;
}




static void zc_out(zcState * state, char ch)
{
	int		n;
	
	zc_cursor(state, HIDE);
	
	switch ((unsigned char)ch) {
		case 0x07:		//bell
			SysBeep(1);
			break;
			
		case 0x08:		// backspace
		case 0x7F:		// delete
			if (state->curCol == 0) {
				if (state->curRow > 0) {
					state->curRow -= 1;
					state->curCol = state->numCols;
				}
				else {
					break;
				}
			}

			state->curCol -= 1;
			MoveTo(
				state->curCol * state->charW + LEFT_MARGIN,
				state->curRow * state->charH + state->charBase + TOP_MARGIN);
			DrawChar(' ');
			(*(state->image))[state->curRow * state->numCols + state->curCol] = ' ';
			break;
			
		case '\t':
			n = 8 - (state->curCol % 8);
			while (n--) zc_out(state, ' ');
			break;

		case '\n':
			state->curCol = 0;
			state->curRow += 1;
			if (state->curRow >= state->numRows) {
				Rect		area;
				
				area.top = TOP_MARGIN;
				area.left = LEFT_MARGIN;
				area.bottom = state->charH * state->curRow + TOP_MARGIN;
				area.right = state->charW * state->numCols + LEFT_MARGIN;
				
				ScrollRect(&area, 0, - state->charH, NULL);
				area.top = area.bottom - state->charH;
				area.bottom += BOTTOM_MARGIN;
				EraseRect(&area);
				
				state->curRow -= 1;
				memmove(*(state->image), 
					*(state->image) + state->numCols,
					(state->numRows - 1) * state->numCols);
				memset(*(state->image) + ((state->numRows - 1) * state->numCols),
					' ', state->numCols);
			}
			break;

		case '\r':
			state->curCol = 0;
			break;

		default:
			MoveTo(
				state->curCol * state->charW + LEFT_MARGIN,
				state->curRow * state->charH + state->charBase + TOP_MARGIN);
			DrawChar(ch);
			(*(state->image))[state->curRow * state->numCols + state->curCol] = ch;
			state->curCol += 1;
			if (state->curCol >= state->numCols)
				zc_out(state, '\n');
			break;
	}
	
	state->cursorBlink = 1;
	(*g)->lastBlink = TickCount();
	zc_cursor(state, SHOW);
}

static void zc_out_buf(zcState * state, const char * string, int nChars)
{
	int		i, j;
	int		lineLeft;
	char	c;
	
	zc_cursor(state, HIDE);
	
	for (i = 0; i < nChars; ) {
		j = i;
		lineLeft = state->numCols - state->curCol;
		
		while (j < nChars
		&&     (j - i) <= lineLeft
		&&     (c = string[j]) != '\0'
		&&     strchr("\7\8\t\n\r\377", c) == NULL)
			j += 1;
		
		j -= i;
		if (j > 0) {
			MoveTo(
				state->curCol * state->charW + LEFT_MARGIN,
				state->curRow * state->charH + state->charBase + TOP_MARGIN);
			DrawText(string, i, j);
			memcpy(*(state->image) + (state->curRow * state->numCols + state->curCol),
				string + i,
				j);
			state->curCol += j;
			if (state->curCol >= state->numCols)
				zc_out(state, '\n');
			i += j;
		}
		else {
			c = string[i];
			if (c == '\0') break;
			zc_out(state, c);
			i += 1;
		}
	}
	
	state->cursorBlink = 1;
	(*g)->lastBlink = TickCount();
	zc_cursor(state, SHOW);
}


static void zc_in(zcState * state, char ch)
{
	int			n;
	
	if (! state->raw) {
		switch ((unsigned char)ch) {
			case 0x08:		//backspace
				if (state->inLine == state->inEnd) return;
				state->inEnd -= 1;
				if (state->echo)
					zc_out(state, ch);
				return;
				
			case 0x15:		//ctrl-u, clear to begining of line
				while (state->inLine != state->inEnd) {
					state->inEnd -= 1;
					if (state->echo)
						zc_out(state, 0x08);
				}
				return;
				
			case 0x17:		//ctrl-w, clear last word
				break;
				
			case '\r':		//new line
				ch = '\n';
				break;
					
			default:
				break;
		}
	}
	
	if (state->inEnd == INPUT_BUFFER_SIZE) {
		//buffer is full, try sliding it left
		n = state->inStart;
		if (n > 0) {
			memmove(state->input, state->input + n, INPUT_BUFFER_SIZE - n);
			state->inStart = 0;
			state->inLine -= n;
			state->inEnd -= n;
		}
		else {
			//buffer is totally full, just drop character...
			SysBeep(1);
			
			if (ch == '\n'  &&  ! state->raw) {
				state->input[state->inEnd - 1] = ch;
				state->inLine = state->inEnd;
			}

			return;
		}
	}
	
	state->input[state->inEnd++] = ch;
	
	if (ch == '\n')
		state->inLine = state->inEnd;
	
	if (state->echo)
		zc_out(state, ch);
}


static void zc_repaint(zcState * state)
{
	char		*image;
	int			i;
	
	HLock(state->image);
	image = *(state->image);
	
	for (i = 0; i < state->numRows; i++) {
		MoveTo(LEFT_MARGIN,
			i * state->charH + state->charBase + TOP_MARGIN);
		DrawText(image,
			i * state->numCols,
			state->numCols);
	}
	
	if (state->growable) {
		ClipRect(&state->growRect);
		DrawGrowIcon(state->window);
		SetClip(state->clipRgn);
	}
	
	zc_cursor(state, PAINT);
	
	HUnlock(state->image);
}



int zcGetc(zcConsole c)
{
	zcState		* state;
	int			nextChar;
	
	state = zc_use(c);
	if (state == NULL) return zcEOF;
	
	while (state->inStart == (state->raw ? state->inEnd : state->inLine)) {
		if (state->delay) {
			zc_done(c);
			(*g)->mgr();
			state = zc_use(c);
			if (state == NULL) return zcEOF;
		}
		else {
			nextChar = zcNoInput;
			goto done;
		}
	}
	
	nextChar = state->input[state->inStart];
	
	if (state->inStart == state->inLine) {
		state->inLine += 1;
	}
	state->inStart += 1;

done:
	if (state->inStart == state->inEnd) {
		state->inStart =
		state->inEnd =
		state->inLine = 0;
	}
	
	zc_done(c);
	return nextChar;
}


int zcGets(zcConsole c, char *string, int nchars)
{
	int		i;
	
	i = zcGetbuf(c, string, nchars - 1);
	
	if (i < zcOK)
		string[0] = 0;
	else
		string[i] = 0;
	
	return i;
}


int zcGetbuf(zcConsole c, char *string, int nchars)
{
	zcState		* state;
	int			i;
	int			nextChar;
	
	state = zc_use(c);
	if (state == NULL) return zcEOF;
	
	for (i = 0; i < nchars; i++) {
		while (state->inStart == (state->raw ? state->inEnd : state->inLine)) {
			if (state->delay) {
				zc_done(c);
				(*g)->mgr();
				state = zc_use(c);
				if (state == NULL)
					return i ? i : zcEOF;
			}
			else {
				i = zcNoInput;
				goto done;
			}
		}
		
		nextChar = state->input[state->inStart];
		if (state->inStart == state->inLine) {
			state->inLine += 1;
		}
		state->inStart += 1;
		
		string[i] = nextChar;
		
		if (nextChar == '\n') {
			i += 1;
			goto done;
		}
	}
	
done:
	if (state->inStart == state->inEnd) {
		state->inStart =
		state->inEnd =
		state->inLine = 0;
	}
			
	zc_done(c);
	return i;
}



int zcPutc(zcConsole c, int ch)
{
	zcState		* state;
	GrafPtr		savePort;
	
	state = zc_use(c);
	if (state == NULL) return zcEOF;

	GetPort(&savePort);
	SetPort(state->window);
	
	zc_out(state, ch);
	
	SetPort(savePort);
	zc_done(c);
	return zcOK;
}

int zcPuts(zcConsole c, const char *string)
{
	zcState		* state;
	GrafPtr		savePort;
	
	state = zc_use(c);
	if (state == NULL) return zcEOF;

	GetPort(&savePort);
	SetPort(state->window);

	zc_out_buf(state, string, (int)(((unsigned)-1)>>1));

	SetPort(savePort);
	zc_done(c);
	return zcOK;
}

int zcPutbuf(zcConsole c, const char *string, int n)
{
	zcState		* state;
	GrafPtr		savePort;
	
	state = zc_use(c);
	if (state == NULL) return zcEOF;

	GetPort(&savePort);
	SetPort(state->window);

	zc_out_buf(state, string, n);
	
	SetPort(savePort);
	zc_done(c);
	return zcOK;
}


int zcTypec(zcConsole c, int ch)
{
	char		buffer[1];
	
	buffer[0] = ch;
	return zcTypebuf(c, buffer, 1);
}

int zcTypes(zcConsole c, const char *string)
{
	return zcTypebuf(c, string, strlen(string));
}

int zcTypebuf(zcConsole c, const char *string, int nchars)
{
	zcState		* state;
	GrafPtr		savePort;
	
	state = zc_use(c);
	if (state == NULL) return zcEOF;

	GetPort(&savePort);
	SetPort(state->window);
	
	while (nchars--)
		zc_in(state, *string++);

	SetPort(savePort);
	zc_done(c);
	return zcOK;
}




void zcFlush(zcConsole c)
{
	int		cStart, cEnd, i;
	zcState	**state;
	
	if (g == NULL)						return;
	if (c == zcNoConsole)				return;
	if (c < 0  ||  c >= MAX_CONSOLES)	return;
	
	if (c == zcAllConsoles)		cStart = 0, cEnd = MAX_CONSOLES;
	else						cStart = cEnd = c;
		
	for (i = cStart; i <= cEnd; i++) {
		state = (*g)->consoles[i];
		if (state != NULL) {
			(*state)->inStart =
			(*state)->inLine =
			(*state)->inEnd = 0;
		}
	}
}


static void zc_handle_blink(void)
{
	WindowPtr		window;
	zcConsole		c;
	zcState			* state;
	GrafPtr			savePort;
	unsigned long	t, lb, ct;

	t = TickCount();
	lb = (*g)->lastBlink;
	ct = GetCaretTime();
	
	if (t >= (lb + ct)) {		
		window = FrontWindow();
		c = zc_find_win(window);
		state = zc_use(c);
		
		if (state != NULL) {
			GetPort(&savePort);
			SetPort(state->window);
						
			zc_cursor(state, HIDE);
			state->cursorBlink = 1 - state->cursorBlink;
			zc_cursor(state, SHOW);
			
			SetPort(savePort);
			zc_done(c);
		}
					
		(*g)->lastBlink = t;
	}
}

int zcHandleEvent(EventRecord *e)
{
	WindowPtr		window;
	zcConsole		c;
	zcState			* state;
	GrafPtr			savePort;
	
	zc_handle_blink();
	
	if (g == NULL) return 0;
	
	switch (e->what) {
		case keyDown:
		case autoKey:
			if (e->modifiers & cmdKey) return 0;
			
			window = FrontWindow();
			c = zc_find_win(window);
			state = zc_use(c);
			if (state == NULL) return 0;
			
			GetPort(&savePort);
			SetPort(state->window);
			
			zc_in(state, e->message & charCodeMask);
			
			SetPort(savePort);
			zc_done(c);
			break;
			
		case mouseDown:
			switch (FindWindow(e->where, &window)) {
				case inContent:
					c = zc_find_win(window);
					if (c == zcNoConsole) return 0;
					SelectWindow(window);
					break;
					
				case inDrag:
					c = zc_find_win(window);
					if (c == zcNoConsole) return 0;
					{
						Rect	dragRect;
						
						dragRect = qd.screenBits.bounds;
						DragWindow(window, e->where, &dragRect);
					}
					break;
				
				case inGoAway:
					c = zc_find_win(window);
					if (c == zcNoConsole) return 0;
					if (TrackGoAway(window, e->where))
						HideWindow(window);
					break;
				
				case inGrow:
					c = zc_find_win(window);
					state = zc_use(c);
					if (state == NULL) return 0;
					{
						Rect	growRect;
						long	result;
						
						growRect.top = 10 * state->charH;
						growRect.bottom = 100 * state->charH;
						growRect.left = 25 * state->charW;
						growRect.right = 120 * state->charW;
						result = GrowWindow(window, e->where, &growRect);
						
						if (result != 0)
							zc_resize(state,
								(HiWord(result) + (state->charH / 2)) / state->charH,
								(LoWord(result) + (state->charW / 2)) / state->charW);
					}
					zc_done(c);
					break;
					
				default:
					return 0;
			}
			break;
			
		case activateEvt:
			window = (WindowPtr)e->message;
			c = zc_find_win(window);
			state = zc_use(c);
			if (state == NULL) return 0;

			GetPort(&savePort);
			SetPort(state->window);
						
			zc_cursor(state, HIDE);
			state->cursorActive = (e->modifiers & activeFlag) != 0;
			state->cursorBlink = 1;
			(*g)->lastBlink = e->when;
			zc_cursor(state, SHOW);
			
			SetPort(savePort);
			zc_done(c);
			break;
			
		case updateEvt:
			window = (WindowPtr)e->message;
			c = zc_find_win(window);
			state = zc_use(c);
			if (state == NULL) return 0;
			
			GetPort(&savePort);
			SetPort(state->window);
			BeginUpdate(window);
						
			zc_repaint(state);
			
			EndUpdate(window);
			SetPort(savePort);
			zc_done(c);
			break;
		
		default:
			return 0;
	}
	
	return 1;
}

void zcHouseKeeping(void)
{
	EventRecord		e;
	
	while (GetNextEvent(everyEvent, &e)) {
		zcHandleEvent(&e);
	}
	
	zc_handle_blink();
}


